"""This module contains utility functions for reading or manipulating entries of PDB files.

:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:date: 2018-03-05
"""

import gzip
import logging
import os
import urllib
from http.client import HTTPResponse
from math import degrees, sqrt
from pathlib import Path
from shutil import copyfileobj
from typing import Iterator, List, Tuple, Dict, Union, TextIO

import Bio.PDB
import numpy as np
from Bio.PDB import Select, PDBList
from Bio.PDB.Chain import Chain
from Bio.PDB.Entity import Entity
from Bio.PDB.Model import Model
from Bio.PDB.PDBIO import PDBIO
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.Residue import Residue
from Bio.PDB.Structure import Structure
from Bio.PDB.vectors import Vector
from Bio.SeqUtils import seq1
from numpy import ndarray
from scipy import spatial

from atligator.geometry import InternalCoordinates, Matrix

logger = logging.getLogger(__name__)

# typedef "forwarded" from module atligator.structure
LigandInfo = Tuple[str, str, str]  # (chain_id, residue_id, new_residue_type)
BinderInfo = Tuple[str, str]  # (chain_id, residue_id)

canonical_amino_acids: List[str] = ['ALA', 'ARG', 'ASN', 'ASP', 'CYS', 'GLN', 'GLU', 'GLY', 'HIS', 'ILE',
                                    'LEU', 'LYS', 'MET', 'PHE', 'PRO', 'SER', 'THR', 'TRP', 'TYR', 'VAL']
canonical_amino_acids_short: List[str] = ['A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G', 'H', 'I',
                                          'L', 'K', 'M', 'F', 'P', 'S', 'T', 'W', 'Y', 'V']
aa_3to1_conv: Dict[str, str] = dict(zip(canonical_amino_acids, canonical_amino_acids_short))
aa_1to3_conv: Dict[str, str] = dict(zip(canonical_amino_acids_short, canonical_amino_acids))

# TODO add alternative names that should be treated as aas:
# https://github.com/EvanBaugh/miscellaneous_scripts/blob/master/process_pdb.py


# source: http://www.imgt.org/IMGTeducation/Aide-memoire/_UK/aminoacids/charge/
# Methionine: Biswal et al.: Sulfuric hydrogen bonds
# TODO pi ion interactions...
sidechain_functions: Dict[Tuple[str, str], List[str]] = {
    ('ARG', 'NE'): ['don'], ('ARG', 'NH1'): ['pos', 'don'], ('ARG', 'NH2'): ['pos', 'don'],
    ('ASN', 'ND2'): ['don'], ('ASN', 'OD1'): ['acc'],
    ('ASP', 'OD1'): ['neg', 'acc'], ('ASP', 'OD2'): ['neg', 'acc'],
    ('GLN', 'NE2'): ['don'], ('GLN', 'OE1'): ['acc'],
    ('GLU', 'OE1'): ['neg', 'acc'], ('GLU', 'OE2'): ['neg', 'acc'],
    ('HIS', 'CG'): ['pi'], ('HIS', 'CD2'): ['pi'], ('HIS', 'CE1'): ['pi'],
    ('HIS', 'ND1'): ['pos', 'don', 'acc', 'pi'], ('HIS', 'NE2'): ['pos', 'don', 'acc', 'pi'],
    ('LYS', 'NZ'): ['pos', 'don'],
    ('MET', 'SD'): ['acc'],
    ('PHE', 'CG'): ['pi'], ('PHE', 'CD1'): ['pi'], ('PHE', 'CD2'): ['pi'],
    ('PHE', 'CE1'): ['pi'], ('PHE', 'CE2'): ['pi'], ('PHE', 'CZ'): ['pi'],
    ('SER', 'OG'): ['don', 'acc'],
    ('THR', 'OG1'): ['don', 'acc'],
    ('TRP', 'NE1'): ['don', 'pi'], ('TRP', 'CG'): ['pi'], ('TRP', 'CD1'): ['pi'], ('TRP', 'CD2'): ['pi'],
    ('TRP', 'CE2'): ['pi'], ('TRP', 'CE3'): ['pi'], ('TRP', 'CZ2'): ['pi'], ('TRP', 'CZ3'): ['pi'],
    ('TRP', 'CH2'): ['pi'],
    ('TYR', 'OH'): ['don', 'acc'], ('TYR', 'CG'): ['pi'], ('TYR', 'CD1'): ['pi'], ('TYR', 'CD2'): ['pi'],
    ('TYR', 'CE1'): ['pi'], ('TYR', 'CE2'): ['pi'], ('TYR', 'CZ'): ['pi'],
    ('ALA', 'O'): ['acc'], ('ALA', 'N'): ['don'], ('ARG', 'O'): ['acc'], ('ARG', 'N'): ['don'],
    ('ASN', 'O'): ['acc'], ('ASN', 'N'): ['don'], ('ASP', 'O'): ['acc'], ('ASP', 'N'): ['don'],
    ('CYS', 'O'): ['acc'], ('CYS', 'N'): ['don'], ('GLN', 'O'): ['acc'], ('GLN', 'N'): ['don'],
    ('GLU', 'O'): ['acc'], ('GLU', 'N'): ['don'], ('GLY', 'O'): ['acc'], ('GLY', 'N'): ['don'],
    ('HIS', 'O'): ['acc'], ('HIS', 'N'): ['don'], ('ILE', 'O'): ['acc'], ('ILE', 'N'): ['don'],
    ('LEU', 'O'): ['acc'], ('LEU', 'N'): ['don'], ('LYS', 'O'): ['acc'], ('LYS', 'N'): ['don'],
    ('MET', 'O'): ['acc'], ('MET', 'N'): ['don'], ('PHE', 'O'): ['acc'], ('PHE', 'N'): ['don'],
    ('PRO', 'O'): ['acc'], ('PRO', 'N'): ['don'], ('SER', 'O'): ['acc'], ('SER', 'N'): ['don'],
    ('THR', 'O'): ['acc'], ('THR', 'N'): ['don'], ('TRP', 'O'): ['acc'], ('TRP', 'N'): ['don'],
    ('TYR', 'O'): ['acc'], ('TYR', 'N'): ['don'], ('VAL', 'O'): ['acc'], ('VAL', 'N'): ['don'],
}

ABC = ("A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M",
       "N", "O", "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z",
       "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m",
       "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z",
       "0", "1", "2", "3", "4", "5", "6", "7", "8", "9")

# Conversion dictionaries for chain processing within ALARMS
chain_conv = dict(zip(ABC, range(len(ABC))))
chain_inv_conv = dict(zip(range(len(ABC)), ABC))


def get_interaction_types(restype1: str, atom1: str, restype2: str, atom2: str) -> List[str or None]:
    """Determines a specific interaction type between two atoms of different residues.
    :param restype1: three letter code of first residue
    :param atom1: atom code for first atom
    :param restype2: three letter code for second residue
    :param atom2: atom code for second atom
    :return: one of ['ionic', 'hbond', 'aromatic', 'unspecific']
    """
    result = []
    try:
        sf1 = sidechain_functions[(restype1, atom1)]
        sf2 = sidechain_functions[(restype2, atom2)]

        if 'pos' in sf1 and 'neg' in sf2:
            result.append('ionic')
        elif 'neg' in sf1 and 'pos' in sf2:
            result.append('ionic')

        if 'don' in sf1 and 'acc' in sf2:
            result.append('hbond')
        elif 'acc' in sf1 and 'don' in sf2:
            result.append('hbond')

        if 'pi' in sf1 and 'pi' in sf2:
            result.append('aromatic')
    except KeyError:
        pass
    result.append(None)

    return result


def _get_child_by_id(parent: Entity, child_id: str) -> Entity or None:
    """
    returns the child entity with specified ID in the PDB parent entity.
    This is not supposed to be executed by itself, rather wrapped by a function to check for the correct entity type.
    Thus, not tested yet.
    :param parent: the PDB entity to search.
    :param child_id: the ID of the child to find
    :return: the retrieved child, or None if not exists
    """
    try:
        return parent.child_dict[child_id]
    except KeyError:
        return None


def get_chain_by_id(model: Model, chain_id: str) -> Chain or None:
    """returns the chain with specified ID in the PDB structure.
    :param model: the PDB model to search
    :param chain_id: the ID of the chain to find
    :return: the retrieved chain, or None if not exists
    """
    if isinstance(model, Model):
        return _get_child_by_id(model, chain_id)
    else:
        raise TypeError


def get_path(residue: Residue) -> str:
    """
    Obtains a unique path name for a PDB residue. Example: '5XYZ/0/A/GLU250'
    :param residue:
    :return: a globally unique path residue
    """
    structure, model, chain, id_ = residue.get_full_id()
    return "/".join((structure, str(model), chain, residue.get_resname() + str(id_[1])))


def get_matching_sequences(model: Model, minlen: int, maxlen: int) -> Iterator[str]:
    """identifies chains matching the specified range in the structure and returns their sequences in IUPAC one letter
    code.
    :param model: the PDB model to search
    :param minlen: lower bound of the search range (inclusive)
    :param maxlen: upper bound of the search range (inclusive)
    :return: a generator for sequential strings
    """
    for chain in model.get_chains():
        s = ""
        aa_residues: List[Residue] = [r for r in chain.get_residues() if is_amino_acid_residue(r)]
        for res in aa_residues:
            s += seq1(res.get_resname())
        length = len(s)
        if minlen <= length <= maxlen:
            yield s


def is_canonical_amino_acid(resname: str) -> bool:
    """
    Determines whether the given name is part of the canonical amino acids
    :param resname: the amino acid type to look for
    :return: True if amino acid is accepted
    """
    return resname in canonical_amino_acids


def has_amino_acid_atoms(residue: Residue, glycine: bool = False) -> bool:
    """
    Determines whether the given residue includes C and CA and N atoms, as well as CB if it's not a glycine.
    :param residue: Bio.PDB residue entity to look for the atoms
    :param glycine: If true CB is not mandatory to pass
    :return: True if atoms are found
    """
    if glycine:
        return 'CA' in residue.child_dict and 'C' in residue.child_dict and 'N' in residue.child_dict
    else:
        return 'CA' in residue.child_dict and 'C' in residue.child_dict and 'N' in residue.child_dict and \
               'CB' in residue.child_dict


def is_amino_acid_residue(residue: Residue, extended: str or None = None) -> bool:
    """determines whether the given is an accepted amino acid residue, i.e.,
    it is part of the 20 canonical amino acids or an amino acid type specified in the extended parameter AND
    it has C and CA and N atoms, as well as CB if it's not a glycine.
    :param residue: the residue to check
    :param extended: if filled with str this string is an accepted aa name
    :return: whether the required properties are fulfilled for residue
    """
    resname = residue.get_resname()

    if is_canonical_amino_acid(resname):
        return has_amino_acid_atoms(residue, glycine=True if resname == "GLY" else False)
    elif extended is not None:
        return resname == extended and has_amino_acid_atoms(residue, glycine=True if resname == "GLY" else False)
    return False


def find_ligands(model: Model, minlen: int, maxlen: int, alternative_lig_aa: str or None = None) -> Iterator[Chain]:
    """returns a list of chains that satisfy the ligand criteria. As a first approximation, we assume that a ligand is
    a chain in the structure that contains >=minlen and <=maxlen residues.
    :param model: where to identify ligands
    :param minlen: minimum length of a chain for being a ligand
    :param maxlen: maximum length of a chain for being a ligand
    :param alternative_lig_aa: If given this alternative name of non-canonical aas will be also searched for as a
    ligand residue
    :return: a generator for chains being considered as ligands
    """
    for chain in model.get_chains():
        aa_residues: List[Residue] = [r for r in chain.get_residues()
                                      if is_amino_acid_residue(r, extended=alternative_lig_aa)]
        if minlen <= len(aa_residues) <= maxlen:
            yield chain


class NoBinderAtoms(Exception):
    """raised if no resdiue atoms are found in the binder chain."""

    def __init__(self, message="The binder chain does not contain suitable atoms!"):
        self.message = message
        super().__init__(self.message)


class NoLigandAtoms(Exception):
    """raised if no no atoms are found in other chains then the binder chain (ligand)."""

    def __init__(self, message="The chains other than the binder chain do not include suitable atoms!"):
        self.message = message
        super().__init__(self.message)


class NoResidueInDistance(Exception):
    """raised if no ligand residue is found in distance of the binder."""

    def __init__(self, message="No ligand residue is found in given distance of binder chain!"):
        self.message = message
        super().__init__(self.message)


def get_nonbinder_residues_within_radius(model: Model, binder_chain: Chain,
                                         ir_max: float = 8.0, ligand_min_len: int = 3,
                                         segmentation: bool = True,
                                         include_hydrogens: bool = True,
                                         check_canonical_aa: bool = False) -> Iterator[Residue]:
    """retrieves residues that are not part of the binder and lie within the interaction radius with any atom of a
    specified binder. Non-amino acid residues are ignored; optionally, hydrogen atoms are ignored.
    :param model: the PDB model from where to retrieve interactions
    :param binder_chain: residues of this chain are ignored as no self-interactions are recorded
    :param ir_max: interaction radius for interaction between arbitrary atoms in Angstrom
    :param segmentation: If True the binder residues get segmented. This allows to use min length of binder segments.
    :param ligand_min_len: Minimum length of segment of ligand chain
    :param include_hydrogens: whether hydrogen atoms shall be considered for all residues
    :param check_canonical_aa: whether to check if included residues are canonical amino acids or not
    :return: a generator of residues identified as interacting partners
    """

    # Define binder atoms: In binder chain and no hydrogens (default)
    binder_atoms = []
    if include_hydrogens:
        for at in binder_chain.get_atoms():
            if at.get_parent().get_id()[0] == " ":
                binder_atoms.append(at.get_coord())
    elif check_canonical_aa:  # checks if resname is canonical and has backbone atoms
        for at in binder_chain.get_atoms():
            if at.element != 'H' and is_amino_acid_residue(at.get_parent()):
                binder_atoms.append(at.get_coord())
    else:  # only checks if hetflag is " "
        for at in binder_chain.get_atoms():
            if at.element != 'H' and at.get_parent().get_id()[0] == " ":
                binder_atoms.append(at.get_coord())
    try:
        binder_atoms = np.concatenate(binder_atoms, axis=0).reshape((-1, 3))
    except ValueError:
        raise NoBinderAtoms

    ligand_atoms = []
    ligand_atoms_res = []
    for ligand_ch in model.get_chains():
        chain_id = ligand_ch.get_id()
        chain_id_int = chain_conv[chain_id]
        if chain_id == binder_chain.get_id():
            continue
        for ligand_res in ligand_ch.get_residues():
            res_id = ligand_res.get_id()
            if res_id[0] != " " or (check_canonical_aa and is_amino_acid_residue(ligand_res)):
                continue
            for ligand_atom in ligand_res.get_atoms():
                if not include_hydrogens and ligand_atom.element == 'H':
                    continue

                # Select coordinates of corresponding atoms
                ligand_atoms.append(ligand_atom.get_coord())

                # Select residue id and chain id (converted to integer)
                ligand_atoms_res.append(res_id[1])
                ligand_atoms_res.append(chain_id_int)
    try:
        ligand_atoms = np.concatenate(ligand_atoms, axis=0).reshape((-1, 3))
    except ValueError:
        raise NoLigandAtoms
    ligand_atoms_res = np.asarray(ligand_atoms_res).reshape((-1, 2))

    # Build KD tree to find smallest distance to any ligand atom
    mytree = spatial.cKDTree(binder_atoms)
    dist, indexes = mytree.query(ligand_atoms)

    # Select those with distance below threshold and save on residue level
    res_in_dist = ligand_atoms_res[dist <= ir_max]
    if res_in_dist.size < 1:
        raise NoResidueInDistance
    # TODO warning if not possible?
    res_in_dist_uni = np.unique(res_in_dist, axis=0)  # Needs numpy >=1.13. If conda doesnt work (depend.) use pip.

    def consecutive(data, stepsize=1) -> List:
        """
        Returns a list of arrays. These arrays contain consecutive residues.
        :param data: Residues
        :param stepsize: to next residues. Should stay one.
        :return: List of Arrays with residues
        """
        return np.split(data, np.where(np.diff(data.flatten()[0::2].astype(int)) != stepsize)[0] + 1)

    # For segmentation
    if segmentation:
        for chain_int in chain_conv.values():
            chain_res_in_dist = np.take(res_in_dist_uni,
                                        np.where(res_in_dist_uni.flatten()[1::2] == chain_int)[0],
                                        axis=0)
            chain_res_in_dist = np.sort(chain_res_in_dist, axis=0)
            if chain_res_in_dist.size > 0:
                chain_res_in_dist = consecutive(chain_res_in_dist)
                for segment in chain_res_in_dist:
                    if segment.size >= ligand_min_len * 2:  # ligand minimal length!!!
                        for res in segment:
                            try:
                                yield model[chain_inv_conv[res[1]]][int(res[0])]
                            except KeyError:
                                pass
    # alternative mode without segmentation:
    else:
        for segment in res_in_dist_uni:
            yield model[chain_inv_conv[segment[1]]][int(segment[0])]


def get_residues_within_radius(model: Model, ligand: Chain, ligand_residue: Residue,
                               ir_default: float = 4.0, ir_hbond: float = 6.0,
                               ir_aromatic: float = 6.0, ir_ionic: float = 8.0,
                               include_hydrogens: bool = True,
                               allow_ligand_residues: bool = False, skip_bb_atoms: bool = False) -> Iterator[Residue]:
    """retrieves residues that are not part of the ligand (if allow_ligand_residues False) and lie within the
    interaction radius with any atom of a specified ligand. Non-amino acid residues are ignored; optionally
    for side-chain interactions, hydrogen atoms are ignored.
    :param model: the PDB model from where to retrieve interactions
    :param ligand: residues of this chain are ignored as no self-interactions are recorded
    :param ligand_residue: the ligand residue which is searched for interactions with non-ligand residues
    :param ir_default: interaction radius for interaction between arbitrary atoms in Angstrom
    :param ir_hbond: interaction radius for hydrogen bond interactions in Angstrom
    :param ir_aromatic: interaction radius for aromatic interactions in Angstrom
    :param ir_ionic: interaction radius for ionic interactions in Angstrom
    :param include_hydrogens: whether hydrogen atoms shall be considered for all residues
    :param allow_ligand_residues: If True also ligand_residues are allowed (not neighbours +- 3 residues!)
    :return: a generator of residues identified as interacting partners
    """
    if not is_amino_acid_residue(ligand_residue):
        pass  # TODO ?
    max_radius = max(ir_ionic, ir_aromatic, ir_hbond, ir_default)
    for chain in model.get_chains():

        # iteration over chains
        if chain.get_id() == ligand.get_id() and not allow_ligand_residues:
            continue
        forbidden = set()
        for binder_residue in chain:

            # iteration over residues in chain
            if allow_ligand_residues:
                if -3 < ligand_residue.get_id()[1] - binder_residue.get_id()[1] < 3:
                    continue

            added = False  # switch to see if residue is already found -> other atoms don't have to be iterated
            if binder_residue.get_id()[1] in forbidden:
                continue
            for binder_atom in binder_residue:

                # iteration over atoms in residue

                # skipping hydrogens or backbone atoms depending on arguments
                if not include_hydrogens and binder_atom.element == 'H':
                    continue
                if skip_bb_atoms and binder_atom.get_id() in ("N", "C", "O"):
                    continue

                for ligand_atom in ligand_residue:

                    # iteration over atoms in ligand residue

                    # skipping hydrogens or backbone atoms depending on arguments
                    if not include_hydrogens and ligand_atom.element == 'H':
                        continue
                    if skip_bb_atoms and ligand_atom.get_id() in ("N", "C", "O"):
                        continue

                    # requesting interaction type of specific atoms
                    itypes = get_interaction_types(ligand_residue.get_resname(), ligand_atom.get_id(),
                                                   binder_residue.get_resname(), binder_atom.get_id())
                    if len(itypes) == 1:
                        radius = ir_default
                    else:
                        radii = []
                        if 'ionic' in itypes:
                            radii.append(ir_ionic)
                        if 'aromatic' in itypes:
                            radii.append(ir_aromatic)
                        if 'hbond' in itypes:
                            radii.append(ir_hbond)
                        radius = max(radii)
                    distance: float = (binder_atom.get_vector() - ligand_atom.get_vector()).norm()
                    # Forbidden residues are neighbouring residues that will definitely be outside of maximum radius
                    # from the current ligand residue.
                    # 28 A are representing 8 A (as the the maximum distance from the most distant atoms of a residue)
                    # for the current ligand residue, the current binder_chain residue, the neighbour of current binder_chain
                    # residue as well as 4 A for the distance between two C-alpha atoms of neighbouring residues.
                    if distance > 28.0 + max_radius:
                        forbidden_neighbours = int((distance - 28.0) / 4)
                        curr_id = binder_residue.get_id()[1]
                        for i in range(forbidden_neighbours):
                            forbidden.add(curr_id + i)
                            forbidden.add(curr_id - i)
                    if distance < radius:
                        yield binder_residue
                        added = True
                        break
                if added:
                    break


def get_cbeta_position(residue: Residue) -> Vector:
    """returns the C_beta vector of this amino acid residue. For non-Gly residues, the real CB is returned. For Gly,
    this is simulated by geometric considerations.
    :param residue: the context residue
    :return: a vector (in absolute coordinates) representing the real or hypothetical C_beta location.
    """
    # 'real' C_beta exists
    if residue.has_id('CB'):
        return residue['CB'].get_vector()
    # C_beta does not exist; simulate it assuming that C, N, H_alpha and the virtual C_beta form a perfect tetrahedron
    # with C_alpha as centroid.
    calpha: Vector = residue['CA'].get_vector()
    co: Vector = residue['C'].get_vector()
    n_: Vector = residue['N'].get_vector()
    cent_co_n: Vector = (co + n_) ** 0.5
    cent_cb_h: Vector = cent_co_n + (calpha - cent_co_n) ** 2.0
    orth1 = (co - calpha).normalized()
    orth2 = (n_ - calpha).normalized()
    dir_cb_h = orth1 ** orth2
    length: float = (cent_co_n - co).norm()
    cbeta = cent_cb_h - dir_cb_h ** length
    return cbeta


def get_icoor_from_pdb_residue(residue: Residue) -> InternalCoordinates:
    """Defines an internal coordinate system relative to a given context PDB residue as follows:
    - the origin is the context C_alpha atom
    - the x-axis is defined by the vector between context C_alpha and C_beta.
    - C_alpha, C_beta, and C lie within the xy-plane, such that the internal z axis denotes the normal of this plane.
    - N has always a negative z coordinate (when considering L amino acids)
    - lengths are preserved, no scaling is applied when compared to the external coordinate system.
    - C points to positive y direction! -> y is now -y, otherwise D-AAs are created
    :param residue: the context residue for the internal coordinate system
    :return: an internal coordinate system with the described properties
    """
    calpha: Vector = residue['CA'].get_vector()
    cbeta: Vector = get_cbeta_position(residue)
    co: Vector = residue['C'].get_vector()
    # basis vectors for the internal coordinate system
    internal_x = (cbeta - calpha).normalized()
    planar = (co - calpha).normalized()
    internal_z = (internal_x ** planar).normalized()
    internal_y = -(internal_x ** internal_z).normalized()
    # create the rotation matrix
    rotation_matrix: Matrix = [
        [internal_x[0], internal_x[1], internal_x[2]],
        [internal_y[0], internal_y[1], internal_y[2]],
        [internal_z[0], internal_z[1], internal_z[2]]]
    # return the internal coordinates. The translation vector equals the external C_alpha vector.
    return InternalCoordinates(calpha, rotation_matrix)


def mutate_residue(residue: Residue, restype: str) -> None:
    """Applies a point mutation to a given residue. All atoms except for 'C', 'CA', 'N', 'O' are removed and the
    residue type is set according to the given value.
    :param residue: the residue to mutate
    :param restype: the new residue type
    """
    for scatom in [atom for atom in residue if atom.id not in ['C', 'CA', 'N', 'O']]:
        residue.detach_child(scatom.id)
    residue.resname = restype


def replace_restypes(structure_file: str, old_type: str, new_type: str) -> None:
    """Replaces old_type residues with new_type residues everywhere in the provided PDB structure
    :param structure_file: PDB filename
    :param old_type: all occurrences of this residue type are to be replaced by new_type
    :param new_type
    """
    parser = PDBParser(QUIET=True)
    structure: Structure = parser.get_structure(structure_file, structure_file)
    for residue in structure.get_residues():
        if residue.resname == old_type:
            residue.resname = new_type
    io = PDBIO()
    io.set_structure(structure)
    io.save(structure_file)


class SecondaryScores:
    # TODO add rmsd to original from pymol_util
    def __init__(self, linearity: float, lin_p_res: float, phi: float, psi: float, phi_dev: float, psi_dev: float,
                 rmsd: float):
        self.linearity = linearity
        self.linearity_p_res = lin_p_res
        self.phi = phi
        self.psi = psi
        self.phi_dev = phi_dev
        self.psi_dev = psi_dev
        self.rmsd = rmsd


def n(vector: Tuple) -> ndarray:
    """
    Returning an numpy vector from a Tuple.
    :param vector: Tuple to convert to a vector
    :return: numpy vector
    """
    return np.array(vector)


def get_angle(a: Tuple, b: Tuple, c: Tuple) -> float or None:
    """
    Calculates an angle between the points a<->b<->c with b being the middle.
    Was tested with threedimensional points.
    :param a: point a with variables in a tuple
    :param b: point b with variables in a tuple
    :param c: point c with variables in a tuple
    :return: The angle in radians
    """
    a: ndarray = n(a)
    b: ndarray = n(b)
    c: ndarray = n(c)
    ba = a - b
    bc = c - b
    if np.linalg.norm(ba) == 0.0 or np.linalg.norm(bc) == 0.0:
        return None
    cosine_angle = np.dot(ba, bc) / (np.linalg.norm(ba) * np.linalg.norm(bc))
    cosine_angle = max((-1, min((1, cosine_angle))))
    angle: float = np.arccos(cosine_angle)
    return angle


def calc_dist_to_line(line_point_a: Tuple, line_point_b: Tuple, point_c: Tuple) -> float or None:
    """
    Calculates the distance of point c to a line crossing point a and b.
    :param line_point_a: coordinates of point a in a tuple.
    :param line_point_b: coordinates of point b in a tuple.
    :param point_c: coordinates of point c whose distance should be measured.
    :return: distance as a float
    """
    alpha = get_angle(line_point_a, line_point_b, point_c)
    if alpha is None:
        return None
    distance: float = np.sin(alpha) * np.linalg.norm((n(line_point_b) - n(point_c)))
    return distance


def get_binder_positions(structure: Structure, binder_info: List[BinderInfo]) -> List[int]:
    """
    Returns a list of absolute positions (starting at 1 for the first residue of the first chain) that match any
    content of the binder_info list (= residues of ligand binding site on binder_chain).
    Example: host consists of chain A, residue ids are 11, 12, 16, 17 of which 12 and 16 are part of the binding site.
    Ligand consists of residues 1, 2, 3 of chain D.
    The function would return [2, 3].
    :param structure: a Bio.PDB structure to identify binder_chain positions in
    :param binder_info: a list of tuples containing chain identifer and chain-relative residue id.
    :return: a list of absolute positions indicating the residues contained in binder_info
    """
    result: List[int] = []
    for i, res in enumerate(structure.get_residues()):
        for bi in binder_info:
            res: Residue
            if res.get_parent().get_id() == bi[0] and res.get_id()[1] == int(bi[1]):
                result.append(i + 1)
                break
    return result


def get_ligand_positions(structure: Structure, ligand_info: List[LigandInfo]) -> List[int]:
    """
    Returns a list of absolute positions (starting at 1 for the first residue of the first chain) that match any
    content of the ligand_info list.
    Example: host consists of chain A, residue ids are 11, 12, 16, 17. Ligand consists of residues 1, 2, 3 of chain D.
    The function would return [5, 6, 7].
    :param structure: a PDB structure to identify ligand positions in
    :param ligand_info: a list of tuples containing chain identifer and chain-relative residue id.
    :return: a list of absolute positions indicating the residues contained in ligand_info
    """
    result: List[int] = []
    for i, res in enumerate(structure.get_residues()):
        for li in ligand_info:
            res: Residue
            if res.get_parent().get_id() == li[0] and res.get_id()[1] == int(li[1]):
                result.append(i + 1)
                break
    return result


def inspect_pdb(pdb_id: str, ligand_info: List[LigandInfo], context_path: str = None) -> SecondaryScores:
    """
    inspects pdb file with name pdb_id.pdb in current directory and returns secondary
    scores like the linearity of the chain(s) other than A (-> peptide ligand) or the phi
    and psi angle averages of these chains.
    :param pdb_id: name of the pdb file without the '.pdb'
    :param ligand_info: contains descriptions of the residues of the ligand
    :param context_path: where it all happens
    :return: a SecondaryScore instance containing several attributes like phi/psi angle averages
    """
    for model in Bio.PDB.PDBParser().get_structure(pdb_id, pdb_id + '.pdb'):
        for chain in model:
            if chain.get_id() != ligand_info[0][0]:
                continue
            line_a: Tuple = chain[int(ligand_info[0][1])]['CA'].coord
            line_b: Tuple = chain[int(ligand_info[-1][1])]['CA'].coord
            linearity = 0.0
            n_res = len(chain)
            for residue in chain:
                for atom in residue:
                    if atom.get_id() in ('CA', 'C', 'N'):
                        if int(residue.get_id()[1]) not in (1, len(ligand_info)):
                            dist = calc_dist_to_line(line_a, line_b, atom.coord)
                            if dist is None:
                                continue
                            linearity += dist
            polypeptides = Bio.PDB.PPBuilder().build_peptides(chain)
            for poly in polypeptides:
                phipsi: List[List[float, float]] = poly.get_phi_psi_list()
                phps_deg: List[List[float]] = []
                for res in phipsi[1:-1]:
                    res_phps: List[float] = []
                    for i in range(2):
                        if res[i] is not None:
                            res_phps.append(degrees(res[i]))
                        else:
                            res_phps.append(0.0)
                    phps_deg.append(res_phps)
                phi_avg = 0.0
                psi_avg = 0.0
                tot = 0
                for i in range(len(phps_deg[1:-1])):
                    phi_avg += phps_deg[i][0]
                    psi_avg += phps_deg[i][1]
                    tot = i + 1
                phi_avg /= tot
                psi_avg /= tot
                phi_dev = 0
                psi_dev = 0
                for i in range(len(phps_deg[1:-1])):
                    phi_dev += (phps_deg[i][0] - phi_avg) ** 2
                    psi_dev += (phps_deg[i][1] - psi_avg) ** 2
                phi_dev = sqrt(phi_dev / tot)
                psi_dev = sqrt(psi_dev / tot)
                rmsd = get_rmsd(('' if context_path is None else context_path + os.sep) + 'scaffold',
                                pdb_id, len(ligand_info))
                return SecondaryScores(linearity, linearity / n_res, phi_avg, psi_avg, phi_dev, psi_dev, rmsd)


def get_rmsd(a_pdb: str, b_pdb: str, last_x: int = None) -> float:
    """
    Returns the rmsd value of the two pdb structures in parameters. If last_x parameter is given,
    Only the rmsd for the last 'last_x' residues is calculated. These last are also not taken into
    account for kabsch rotation-translation (very different positions would affect translation and rotation).

    :param a_pdb: Pdb id of first structure in curr. dir. '.pdb' will be added automatically.
    :param b_pdb: Pdb id of second structure in curr. dir. '.pdb' will be added automatically.
    :param last_x: Number of the last residues (peptide ligand?) to consider for RMSD
    :return: rmsd value
    """
    a = np.array(get_pdb_coors(a_pdb))
    b = np.array(get_pdb_coors(b_pdb))
    # Check if both have the same number of res
    assert len(a) == len(b)
    if last_x is None:
        last_x = len(b)
    b -= get_centroid(b[:-last_x])
    a -= get_centroid(a[:-last_x])
    a = rotate_kabsch(a, b, last_x)
    return calc_rmsd(a[-last_x:], b[-last_x:])


def get_pdb_coors(pdb_id: str) -> List[ndarray]:
    for model in Bio.PDB.PDBParser().get_structure(pdb_id, pdb_id + '.pdb'):
        coor_lst = []
        # TODO does it always start with chain A??
        for chain in model:
            for residue in chain:
                for atom in residue:
                    if atom.get_id() == 'CA':  # OR in ('CA', 'C', 'N'):
                        coor_lst.append(atom.coord)
        return coor_lst


def rotate_kabsch(a: ndarray, b: ndarray, not_include: int):
    r = get_u_kabsch(a[:-not_include], b[:-not_include])
    a = np.matmul(a, r)
    return a


def get_u_kabsch(a: ndarray, b: ndarray):
    """
    Returns the rotational matrix for rotation of a to fit b best. Calculated using
    kabsch algorithm.
    :param a: Matrix a including coordinates of first structure
    :param b: Matrix b including coordinates of second structure
    :return: Rotation matrix
    """
    c = np.matmul(np.transpose(a), b)
    # TODO check svd names and transpositons
    v, s, w = np.linalg.svd(c)
    d = (np.linalg.det(v) * np.linalg.det(w)) < 0.0

    if d:
        s[-1] = -s[-1]
        v[:, -1] = -v[:, -1]

    r = np.matmul(v, w)
    return r


def get_centroid(matrix: ndarray) -> ndarray:
    """
    Returning the x centroid values for a matrix with points in a x-dim. coordinate system.
    :param matrix: Matrix with several points and their coordinates
    :return: Centroid values for each dimension as a ndarray
    """
    centroid: ndarray = matrix.mean(axis=0)
    return centroid


def calc_rmsd(a: ndarray, b: ndarray) -> float:
    """
    Calculates the root mean square deviation for pairs of points with x coordinates (3 in 3D coor system) each.
    These points are given by ndarray matrices (a and b).
    :param a: Matrix a including coordinates of first structure
    :param b: Matrix b including coordinates of second structure
    :return: rmsd value
    """
    dimensions = len(a[0])
    n_points = len(a)
    rmsd = 0.0
    for a_i, b_i in zip(a, b):
        rmsd += sum([(a_i[i] - b_i[i]) ** 2.0 for i in range(dimensions)])
    return np.sqrt(rmsd / n_points)


def find_pdb_keys_in_scop_dir_cla(query: str, scop_dir_cla_file: Union[TextIO, HTTPResponse]) -> Union[List[str], None]:
    """
    Parses a SCOP(e) directory classification file (dir.cla.scop*.txt) in order to find a list of PDB keys that match
    the given scop query (e.g., 'a', 'a.118.1', etc.)
    :param query: the class, fold, superfamily, or family to search
    :param scop_dir_cla_file: part of the SCOP(e) database that contains the classification
    :return: list of PDB files that match the query, or None if the query was not well-formed
    """
    pdb_keys = set()
    query_split = query.split('.')
    if len(query_split) < 1:
        return None
    query_split = query_split if len(query_split[-1]) else query_split[:-1]
    decode = True if isinstance(scop_dir_cla_file, HTTPResponse) else False
    for line in scop_dir_cla_file:
        if decode:
            line = line.decode("utf-8")
        entries = line.split()
        if not line.startswith('#') and len(entries) > 5:
            pdb_key = entries[1]
            scop_key = entries[3].split('.')
            add = True
            for i, sk in enumerate(scop_key):
                if i < len(query_split) and sk != query_split[i]:
                    add = False
                    break
            if add and pdb_key not in pdb_keys:
                pdb_keys.add(pdb_key)
    return sorted(list(pdb_keys))


def get_available_scop_dir_cla_file(scop_dir_cla_file: str = None) -> Union[TextIO, HTTPResponse]:
    """
    Returns the opened scop_dir_cla_file necessary for matching scop ids and pdb ids

    :param scop_dir_cla_file: file of the SCOP(e) database that contains the classification. Either string with path to
    the file locally or None. If None the file is downloaded from
    https://scop.berkeley.edu/downloads/parse/dir.cla.scope.2.08-stable.txt
    :return: The opened scop_dir_cla_file
    """
    if scop_dir_cla_file is not None and os.path.isfile(scop_dir_cla_file):
        return open(scop_dir_cla_file, "r")
    logging.info("Downloading https://scop.berkeley.edu/downloads/parse/dir.cla.scope.2.08-stable.txt to search for "
                 "matching patterns...")
    return urllib.request.urlopen("https://scop.berkeley.edu/downloads/parse/dir.cla.scope.2.08-stable.txt")


def find_scop_i_from_pdbs(query: List[str], scop_dir_cla_file: str = None) \
        -> Dict[str, str] or None:
    """
    Parses a SCOP(e) directory classification file (dir.cla.scop*.txt) in order to find the scop identifiers for
    a list of PDB keys.
    :param query: a list of pdb ids you want the scop identifiers from
    :param scop_dir_cla_file: file of the SCOP(e) database that contains the classification. Either string with path to
    the file locally or None. If None the file is downloaded from
    https://scop.berkeley.edu/downloads/parse/dir.cla.scope.2.08-stable.txt
    :return: list of PDB files that match the query, or None if the query was not well-formed
    """
    pdb_keys = set(query)
    pdb_scop_dict = dict((a, None) for a in query)
    with get_available_scop_dir_cla_file(scop_dir_cla_file=scop_dir_cla_file) as fo:
        decode = True if isinstance(scop_dir_cla_file, HTTPResponse) else False
        for line in fo:
            if decode:
                line = line.decode("utf-8")
            entries = line.split()
            if not line.startswith('#') and len(entries) > 5:
                pdb_key = entries[1]
                if pdb_key in pdb_keys:
                    if pdb_scop_dict[pdb_key] is None:
                        scop_key = entries[3]
                        pdb_scop_dict[pdb_key] = [scop_key]
                    elif entries[3] not in pdb_scop_dict[pdb_key]:
                        scop_key = entries[3]
                        logging.info(f"Pdb file found twice: {pdb_key} in {pdb_scop_dict[pdb_key]} and {entries[3]}!")
                        pdb_scop_dict[pdb_key] += [scop_key]
    return pdb_scop_dict


def get_pdb_structure(pdb_key: str, dest_dir: str, local: bool = False, pdb_db_path: str = None) -> str or None:
    """
    Downloads a PDB file matching a given PDB key from a local or remote PDB database. Copies the file into a given
    destination directory and returns the path of the output file.
    If the pdb_key cannot be found in the local database (if local==True) searches in the global PDB database
    (via PDB.Bio.PDBList). If the corresponding pdb_key cannot be found in the database the returned path is not
    touched and can be checked if exists.

    :param pdb_key: The PDB key to find, e.g., 5AEI
    :param dest_dir: file will be copied into this folder as [pdb_key].pdb. If the path does not exist it will be
    created automatically.
    :param local: If True try to copy the file from local copy of pdb db (must be given in pdb_db_path and
    should rescolve : f"{pdb_db_path}/data/{pdb_key[1:3]}/pdb{pdb_key}.ent.gz"
    :param pdb_db_path: Directory that contains a copy of the PDB database. Must contain a folder 'data'.
    :return: the file name, if copying was successful
    """

    pdb_key = pdb_key.lower()

    def return_remote():
        return PDBList().retrieve_pdb_file(pdb_key,
                                           pdir=dest_dir,
                                           file_format="pdb")

    if not local or pdb_db_path is None:
        pdb_path = return_remote()

    else:
        arch = f"{pdb_db_path}/data/{pdb_key[1:3]}/pdb{pdb_key}.ent.gz"
        pdb_path = f"{dest_dir}/{pdb_key}.pdb"
        if not os.path.exists(arch):
            pdb_path = return_remote()
        else:
            with gzip.open(arch, 'rb') as f_in, open(pdb_path, 'wb') as f_out:
                copyfileobj(f_in, f_out)
    return pdb_path


def get_structures_matching_scop_query(query: str, scop_dir_cla_file: str = None):
    """
    Finds all PDB files that match a given SCOP(e) query using a local copy of the SCOP(e) database
    (if scop_dir_cla_file is available) or

    :param query: the class, fold, superfamily, or family to search, e.g., 'a.118.1'
    :param scop_dir_cla_file: file of the SCOP(e) database that contains the classification. Either string with path to
    the file locally or None. If None the file is downloaded from
    https://scop.berkeley.edu/downloads/parse/dir.cla.scope.2.08-stable.txt
    """
    with get_available_scop_dir_cla_file(scop_dir_cla_file=scop_dir_cla_file) as scop_dir_cla_opened:
        return find_pdb_keys_in_scop_dir_cla(query=query, scop_dir_cla_file=scop_dir_cla_opened)


def download_pdbs_by_scop_query(query: str, dest_dir: str, scop_dir_cla_file: str = None, local_db: bool = False,
                                pdb_db_path: str = None) -> None:
    """
    Finds all PDB files that match a given SCOP(e) query using a local copy of the SCOP(e) database, downloads the
    structures from a local copy of the PDB database, and stores the files in PDB format under the given destination
    directory.
    :param query: the class, fold, superfamily, or family to search, e.g., 'a.118.1'
    :param dest_dir: PDB structure files will be copied into this folder
    :param scop_dir_cla_file: part of the SCOP(e) database that contains the classification
    :param local_db: if True it searches for the file in a local copy of the protein database. Must be given in
    pdb_db_path
    :param pdb_db_path: Directory that contains a copy of the PDB database. Must contain a folder 'data'.
    """
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)
    pdbkeys = get_structures_matching_scop_query(query=query, scop_dir_cla_file=scop_dir_cla_file)
    if not len(pdbkeys):
        logger.warning("Found no matching files.")
    else:
        logger.info(f"Found {len(pdbkeys)} matching pdb files.")
    for k in pdbkeys:
        out_file = get_pdb_structure(pdb_key=k, dest_dir=dest_dir, pdb_db_path=pdb_db_path, local=local_db)
        if not os.path.isfile(out_file):
            logging.warning(f"The file {out_file} cannot be found in the pdb database.")
        else:
            logger.info(f'Downloading {out_file} and renaming to '
                        f'{Path(out_file).parent / Path(out_file).name.replace("pdb", "").replace(".ent", ".pdb")}')
            Path(out_file).rename(
                Path(out_file).parent / Path(out_file).name.replace("pdb", "").replace(".ent", ".pdb"))


class ChainSelect(Select):
    """
    A Select class to enable selecting the residues that should be written in an output structure file within
    Bio.PDB.
    """

    # TODO insert a 'remove waters and other non-aa-res ligands' option
    def __init__(self, allowed, chain_id="x"):
        super().__init__()
        self.chain_id = chain_id
        self.allowed = allowed

    def accept_residue(self, residue):
        """
        Only accept residues that are in self.allowed or that belong to the chain in defined self.chain_id (while being
        an amino acid residue)
        :param residue: the residue that get's tested
        :return: bool if this residue should be accepted or not
        """

        return residue in self.allowed or \
               (residue.get_parent().get_id() == self.chain_id and is_amino_acid_residue(residue))



"""This module contains facilities for deep learning atlases and predicting pmatrices from acomplexes based on a
convolutional neural network implemented by SciKit Learn.

:Deprecated: This module has been deprecated because machine learning did not show any improvement compared to the
traditional atlas solution, which implements in fact a linear classification with deterministic parametrization.

:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:date: 2018-03-14
"""

from pickle import dump, load
from typing import List, Dict, Set, Tuple

from Bio.PDB.vectors import Vector

from atligator.acomplex import LigandBinderComplex
from atligator.atlas import Atlas
from atligator.pdb_util import canonical_amino_acids
from atligator.prediction_matrix import BinderPredictionMatrix, BinderPredictionLine, BinderPredictionColumn
from atligator.structure import DesignedBinderResidue, DesignedLigandResidue, get_icoor_from_aresidue


def get_labels_from_resname(restype: str) -> List[float]:
    """Converts a residue type into a feature vector of length 20, where all positions contain the defined negativity
    value, 0.0, except for the index of the residue in the canonical list, which will contain the positivity value 1.0.
    :param restype: three letter code of the residue type to encode
    :returns: a vector of floats with length 20 that represents the feature vector that can be fed into a classifier
    """
    vec = [0.0] * len(canonical_amino_acids)
    try:
        index = canonical_amino_acids.index(restype)
        vec[index] = 1.0
    except ValueError:
        pass
    return vec


def get_labels_from_vector(vector: Vector, scale: float = 1.0) -> List[float]:
    """Scales a vector in order to better fit into an MLP classifier.
    :param vector: the vector to scale. Coordinates in Angstrom
    :param scale: the scalar value to multiply the vector with. If unset, no scaling is applied (factor 1.0)
    :return: a list with three floats that represent the scaled vector that can be fed to MLP
    """
    return (vector ** scale).get_array().tolist()


class DeepAtlas:
    """Instances of this class hold an MLP classifier incorporating the training data extracted from an atlas. The
    underlying classification problem is understood as follows:
    Ligand Residue Type x Ligand C_alpha x Ligand C_alpha->C_beta x Ligand C_alpha->C_O ==> Binder Residue Type
    The topology of the neural network is as follows:
    Input Layer: 29 neurons (20 for ligand residue type label, 3 each for binder C_alpha, orientation and secor coors)
    Hidden Layer: 25 neurons
    Output Layer 20 neurons (representing the possible values for the binder residue type label)"""

    def __init__(self, classifier=None, max_distance: float = 8.0, hidden_neurons: int = 25,
                 activation: str = 'tanh', learning_rate: float = 0.005, max_iter: int = 500, alpha: float = 0.01,
                 tolerance: float = 1e-6, verbose_training: bool = False):
        """Creates a new instance.
        :param classifier: if set, this will be the classifier of the deep atlas. Otherwise, a new MLP classifier is
        created based on experience parameters.
        :param max_distance: the maximum distance between C_alpha atoms in the underlying atlas (used for scaling
        the input vectors)
        :param hidden_neurons: size of the hidden layer of the underlying neural classificator. According to Jeff
        Heaton's rule of thumb, pick a layer size that is the mean of input and output neurons. Therefore, the default
        is 25.
        :param activation: the activation function to use for backpropagation. The default, tanh, works well with
        continuous inputs between -1.0 and 1.0, but also with discrete inputs between 0.0 and 1.0
        :param learning_rate: how fast (but also how biased) learning will happen
        :param max_iter: maximum number of iterations per training
        :param alpha: this factor trains the sharpness of the output
        :param tolerance: learning is stopped when training loss is below this value
        :param verbose_training: whether to pass over klearn's console output
        """
        if classifier is not None:
            self.classifier = classifier
        else:
            from sklearn.neural_network import MLPClassifier
            self.classifier = MLPClassifier(
                hidden_layer_sizes=(hidden_neurons,),
                activation=activation,
                learning_rate_init=learning_rate,
                max_iter=max_iter,
                alpha=alpha,
                tol=tolerance,
                verbose=verbose_training
            )
        self.max_distance = max_distance

    def train(self, atlas: Atlas) -> None:
        """Trains the classifier with input data originating from an atlas.
        :param atlas: its datapoints will be used as training samples.
        """
        data: List[List[float]] = []
        targets: List[List[float]] = []
        for ares in atlas.datapoints:
            sample = get_labels_from_resname(ares.ligand_restype)
            sample.extend(get_labels_from_vector(ares.calpha, 1.0 / self.max_distance))
            sample.extend(get_labels_from_vector(ares.orientation().normalized()))
            sample.extend(get_labels_from_vector(ares.secondary_orientation().normalized()))
            data.append(sample)
            targets.append(get_labels_from_resname(ares.binder_restype))
        self.classifier.fit(data, targets)

    def predict(self, binder_res: DesignedBinderResidue, ligand_res: DesignedLigandResidue,
                prob_threshold: float = 0.001) -> Dict[str, float]:
        """Using the underlying classificator, this method predicts the residue type of a of a given binder residue,
        relative to a given ligand residue, to whose interal coordinate system all binder vectors are transformed first.
        :param binder_res: the binder residue whose residue type to predict
        :param ligand_res: the context ligand residue.
        :param prob_threshold: classification results with probabilities below this threshold will be discarded
        :param: a dictionary that maps significantly probable residue types to the predicted probability
        """
        icoor = get_icoor_from_aresidue(ligand_res)
        sample = get_labels_from_resname(ligand_res.restype)
        sample.extend(get_labels_from_vector(icoor.external_to_internal(
            binder_res.calpha(), False), 1.0 / self.max_distance))
        sample.extend(get_labels_from_vector(icoor.external_to_internal(
            binder_res.orientation(), True).normalized()))
        sample.extend(get_labels_from_vector(icoor.external_to_internal(
            binder_res.secondary_orientation(), True).normalized()))
        prediction: List[float] = self.classifier.predict_proba([sample])[0]
        result: Dict[str, float] = {}
        for i, v in enumerate(prediction):
            restype: str = canonical_amino_acids[i]
            value = prediction[i]
            if value > prob_threshold:
                result[restype] = float(value)
        return result

    def save(self, filename) -> None:
        """Saves the classificator in binary format in the specified file.
        :param filename: where to serialize"""
        with open(filename, 'wb') as fo:
            dump(self, fo)


def load_deep_atlas(filename: str) -> DeepAtlas:
    """Creates a new atlas based on a serialized classificator instance.
    :param filename: from where to load the classificator
    :return: a corresponding deep atlas
    """
    with open(filename, 'rb') as fo:
        deep_atlas: DeepAtlas = load(fo)
    return deep_atlas


def generate_deep_prediction_matrix(acomplex: LigandBinderComplex, deep_atlas: DeepAtlas, distance_cutoff: float = 8.0,
                                    distance_exponent: float = 2.0) -> BinderPredictionMatrix:
    """Predicts possible residue types of the designed residues of a given ligand binder complex based on a deep atlas.
    To this end, every pair of designed ligand and binder residue, which lie within a given cutoff radius, is fed to
    the classificator. The local predictions of all ligand residues are superimposed and stored in the resulting
    binder prediction matrix
    :param acomplex: the ligand binder complex to create the matrix for
    :param deep_atlas: the deep atlas that is used for classification-based prediction
    :param distance_cutoff: maximum distance between ligand and binder C_alpha atoms
    :param distance_exponent: The relative distance (between 0.0 and 1.0) will be raised with this value before being
            multiplied to the score. Therefore, this exponent denotes the "steepness" of distance influence
    :return: a binder prediction matrix calculated based on the classification as explained
    """
    lines: List[BinderPredictionLine] = []
    for binder_res in acomplex.binder.residues:
        columns_map: Dict[str, Tuple[str, Set[int], float]] = {}
        for ligrespos, ligand_res in enumerate(acomplex.ligand.residues):
            distance: float = (ligand_res.calpha - binder_res.calpha).norm()
            if distance < distance_cutoff:
                prediction = deep_atlas.predict(binder_res, ligand_res)
                for pred_restype, pred_score in prediction.items():
                    score = float(pred_score) * (((distance_cutoff - distance) / distance_cutoff) ** distance_exponent)
                    if pred_restype not in columns_map:
                        columns_map[pred_restype] = (pred_restype, set(), 0.0)
                    col = columns_map[pred_restype]
                    columns_map[pred_restype] = (col[0], col[1].union({ligrespos}), col[2]+score)
        columns: List[BinderPredictionColumn] = []
        for col in columns_map.values():
            columns.append(BinderPredictionColumn(col[0], len(col[1]), col[2]))
        lines.append(BinderPredictionLine(binder_res.residue_id, binder_res.original_residue_type, columns))
    return BinderPredictionMatrix(lines)



"""This module provides an interface for different scoring functions that can be used, e.g., with the genetic mutator.

:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:date: 2018-04-19
"""

import random
from shutil import copyfile

from atligator.acomplex import ComplexDescriptor


class ScoreResult:
    """Instances of this class represent the result of a scoring done by a subclass of BaseScorer."""

    def __init__(self, score: float, optimized_structure: str, **kwargs):
        """Creates a new instance.
        :param score: the score calculated
        :param optimized_structure: the conformation that represents the solution with the best score
        :param kwargs: additional properties to assign to this score result
        """
        self.score = score
        self.optimized_structure = optimized_structure
        self.properties = kwargs

    def __getattr__(self, item):
        return self.properties[item]

    def __str__(self):
        return str(self.score) + ' [' + ', '.join([str(k) + ': ' + str(v) for k, v in self.properties.items()]) + ']'

    def __repr__(self):
        return str(self)


class BaseScorer:
    """This abstract base class provides an interface for different scorers. Concrete scorers must implement the
    score function."""

    def score(self, structure: str, descriptor: ComplexDescriptor) -> ScoreResult:
        """To be implemented by subclasses. Evaluates the structure and returns the resulting score and optimized
        structure
        :param structure: the structure to optimize
        :param descriptor: defines ligand and binder residues of the structure
        :return: the corresponding score result
        """
        raise NotImplementedError


class RandomScorer(BaseScorer):
    """This scorer returns a random value between -1000.0 and 0.0. Use it for testing purposes only."""

    def score(self, structure: str, descriptor: ComplexDescriptor) -> ScoreResult:
        relaxed_file = structure.replace(".pdb", "_0001.pdb")
        copyfile(structure, relaxed_file)
        score = random.uniform(-1000.0, 0.0)
        return ScoreResult(score, relaxed_file)



"""This module abstracts from the Rosetta program suite by wrapping the interface of the relax application.

:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:date: 2018-04-04
"""

import os
import shutil
from glob import glob
from subprocess import run, CalledProcessError
from typing import List, Any

from Bio.PDB.PDBParser import PDBParser
from numpy import std

from atligator.acomplex import ComplexDescriptor
from atligator.base_scorer import BaseScorer, ScoreResult
from atligator.pdb_util import SecondaryScores, inspect_pdb, get_ligand_positions, get_binder_positions


def run_command(rosetta_path: str, exe_file: str, db_path: str, input_file: str, output_dir: str, score_file: str,
                extra_rotamers: bool, mode: str, nstruct: int, script: str or None, stdout: Any,
                resfile: str = "./resfile.res") -> None:
    command = f"{rosetta_path}{os.sep}{exe_file} -database {db_path}" + \
              f" -in:file:s {input_file} -out:path:all {output_dir}" + \
              f" -out:path:score {output_dir} -out:file:scorefile {score_file}" + \
              (" -ex1 -ex2" if extra_rotamers else "")
    if mode == "fixbb":
        command += f" -resfile {resfile} -minimize_sidechains"
    elif mode == "score":
        command += f" -out:nstruct {nstruct}" + (" -relax:fast" if script is None else f" -relax:script {script}")
    try:
        run(command, shell=True, check=True, stdout=stdout)
    except CalledProcessError as err:
        print(f"Unexpected {err=}, {type(err)=}")
        raise


class RosettaScorer(BaseScorer):
    """This class encapsulates Rosetta's relax functionality and provides a minimal scripting interface."""

    def __init__(self, rosetta_path: str = '$ROSETTA', relax_executable: str = 'relax.default.linuxgccrelease',
                 rosetta_db_path: str = '$ROSETTA_DB', mode: str = 'relax', relax_script: str = None,
                 extra_rotamers: bool = False, nstruct: int = 1, ligand_factor: float = 20, lin_factor: float = 30,
                 pen_factor: float = 10.0, rmsd_slope: float = 5, rmsd_offset: float = 5, mutant_dir: str = './',
                 binding_site_factor: float = None):
        """Creates a new instance based on information about the Rosetta executable
        :param rosetta_path: the path to the folder where Rosetta's executable binaries are located
        :param relax_executable: the complete name of the relax executable file (including extensions)
        :param rosetta_db_path: the path to the Rosetta databases
        :param mode: Rosetta structure relaxation mode (one of 'score', 'relax')
        :param relax_script: path to a Rosetta relax script, or None if default protocol shall be kept
        :param extra_rotamers: if true, extra rotamers (ex1, ex2) are used
        :param nstruct: the number of structures to produce in every relax run
        :param ligand_factor: the share of ligand energy in the calculation of total energy
        :param lin_factor: linearity impact in total energy calculation
        :param pen_factor: impact of RMSD penalty
        :param rmsd_slope: The slope of the influence of ligand rmsd after the offset
        :param rmsd_offset: An offset value where rmsd penalty will start
        :param mutant_dir: directory where mutant files are stored - relative to current directory.
        :param binding_site_factor: the share of binding site energy in the calculation of total energy
        """
        self.rosetta_path = rosetta_path
        self.relax_executable = relax_executable
        self.rosetta_db_path = rosetta_db_path
        self.mode = mode
        self.relax_script = relax_script
        self.extra_rotamers = extra_rotamers
        self.nstruct = nstruct
        self.ligand_factor = ligand_factor
        self.binding_site_factor = binding_site_factor if binding_site_factor is not None else ligand_factor
        self.lin_factor = lin_factor
        self.pen_factor = pen_factor
        self.rmsd_slope = rmsd_slope
        self.rmsd_offset = rmsd_offset
        self.mutant_dir = mutant_dir if mutant_dir.endswith('/') else mutant_dir + '/'

    def read_score(self, scorefile_path: str, descriptor: ComplexDescriptor, div_stddev: bool = False) -> ScoreResult:
        """Creates a ScoreResult instance from a score file created by Rosetta.
        :param scorefile_path: the name of the scorefile
        :param descriptor: forwarded complex descriptor
        :param div_stddev: If enabled ligand score will be divided by the standard deviation of indiv. residue scores.
        This is not used atm. (see score() return object).
        :return: the ScoreResult runtime object containing both parsed Rosetta scores and secondary scores
        """
        with open(scorefile_path, 'r') as sf:
            line = ""
            while not line.startswith("SCORE:"):
                line = sf.readline()  # ignore headers (if present)
            best_score = float("inf")
            best_binder_score = float("inf")
            best_ligand_score = float("inf")
            best_binding_site_score = float("inf")
            best_lin_score = float("inf")
            best_rmsd = float("inf")
            best_id = ""
            line = sf.readline()
            while not line == "":
                values = line.split()
                total_score = float(values[1])
                mutant_id = self.mutant_dir + values[-1]
                lig_score: float = 0.0
                bin_site_score: float = 0.0
                std_dev_lst: List[float] = []
                parser = PDBParser(QUIET=True)
                mutant_structure = mutant_id + '.pdb'
                biopdb_structure = parser.get_structure(mutant_structure, mutant_structure)
                ligand_positions = get_ligand_positions(biopdb_structure,
                                                        descriptor.ligand_info)
                binding_site_positions = get_binder_positions(biopdb_structure,
                                                              descriptor.binder_info)
                with open(mutant_structure, 'r') as rf:
                    for line in rf.readlines():
                        if any("_" + str(pos) in line for pos in ligand_positions):
                            std_dev_lst.append(float(line.split()[-1]))
                            lig_score += float(line.split()[-1])
                        elif any("_" + str(pos) in line for pos in binding_site_positions):
                            bin_site_score += float(line.split()[-1])
                secondary_scores: SecondaryScores = inspect_pdb(mutant_id, descriptor.ligand_info)
                lin_score = secondary_scores.linearity_p_res
                if div_stddev:
                    lig_score = lig_score / std(std_dev_lst)
                rmsd = secondary_scores.rmsd
                pen = (rmsd - self.rmsd_offset) ** self.rmsd_slope if rmsd >= self.rmsd_offset else 0.0
                # binding site score must be normalized to fit influence of ligand score
                norm_bin_site_score = bin_site_score * len(descriptor.ligand_info) / len(descriptor.binder_info)
                fitness_score: float = total_score + self.ligand_factor * lig_score + self.lin_factor * lin_score + \
                                       self.pen_factor * pen + self.ligand_factor * norm_bin_site_score
                if fitness_score < best_score:
                    best_score = fitness_score
                    best_binder_score = total_score - lig_score
                    best_ligand_score = lig_score
                    best_binding_site_score = bin_site_score
                    best_lin_score = lin_score
                    best_rmsd = rmsd
                    best_id = mutant_id
                line = sf.readline()
        return ScoreResult(best_score,
                           best_id + '.pdb',
                           binder_score=best_binder_score,
                           ligand_score=best_ligand_score,
                           binding_site_score=best_binding_site_score,
                           lin_score=best_lin_score,
                           rmsd_score=best_rmsd)

    def score(self, structure_file: str, descriptor: ComplexDescriptor, score_file_: str = None, log_file: str = None,
              delete_existing_results: bool = True) -> ScoreResult:
        """Invokes Rosetta relax in a shell, waits for the termination of the program and returns the result.
        :param structure_file: the path of a PDB file to relax
        :param descriptor: defines ligand and binder residues
        :param score_file_: path to a file where the output is stored
        :param log_file: Rosetta's command line output is directed into this file
        :param delete_existing_results: whether structures originating from previous Rosetta runs (probably based on
        entirely different mutation sets due to randomization!) shall be removed
        :return: the output of the relax, including the total score
        """
        if score_file_ is None:
            score_file_ = structure_file.lower().replace(".pdb", ".score")
        if log_file is None:
            log_file = structure_file.lower().replace(".pdb", ".log")
        output_path = self.mutant_dir
        with open(log_file, 'w') as lf:
            if delete_existing_results:
                for f in glob(structure_file.replace('.pdb', '_*.pdb')):
                    os.remove(f)
                sf = structure_file.replace('.pdb', '.score')
                if os.path.exists(sf):
                    os.remove(sf)
            executable = self.relax_executable
            if self.mode in ('score', 'fixbb'):
                executable = executable.replace('relax', self.mode)
            run_command(rosetta_path=self.rosetta_path, exe_file=executable, db_path=self.rosetta_db_path,
                        input_file=structure_file, output_dir=output_path, score_file=score_file_,
                        extra_rotamers=self.extra_rotamers, mode=self.mode, nstruct=self.nstruct,
                        script=self.relax_script, stdout=lf)
        if self.mode == 'score':
            shutil.copyfile(f"{structure_file}", f"{structure_file.replace('.pdb', '_0001.pdb')}")
        return self.read_score(score_file_, descriptor)

    def switch_to_nd(self, back: bool = False):
        """
        Alters mutant directory in negative design mode (and back)
        :param back: Reverse to normal mode.
        :return: None
        """
        if not back:
            with open("resfile.res", "w") as rf:
                rf.write("NATAA\nstart")
            self.mutant_dir = self.mutant_dir + "negative_design/"
            self.mode = 'fixbb'
            self.extra_rotamers = True
        else:
            if os.path.exists("./resfile.res"):
                os.remove("./resfile.res")
            self.mutant_dir = self.mutant_dir.split("negative_design/")[0]
            self.mode = 'relax'



"""This module contains classes and functions for the automatic or manual grafting of pockets based on data obtained by
the pocket miner and user-defined input (ligand binder complex descriptors). When several residue pockets are grafted,
the algorithms take care that the globally optimal graft is preferred. All in all, the methods implemented here are very
coarse-grained and the PDB structures generated are physically improper and demand for manual revision and/or
computational optimization.

:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:date: 2018-06-11
"""
import copy
from itertools import product, combinations
from operator import itemgetter
from typing import Callable, Union
from typing import List, Tuple, Dict, Set

from Bio.PDB import PDBParser, PDBIO
from Bio.PDB.Atom import Atom
from Bio.PDB.Residue import Residue
from Bio.PDB.Structure import Structure

from atligator.acomplex import LigandBinderComplex, ComplexDescriptor, generate_complex
from atligator.atlas import AtlasDatapoint
from atligator.pocket_miner import Pocket, PocketCluster
from atligator.structure import DesignedBinderResidue, get_icoor_from_aresidue, DesignedLigandResidue, AbstractResidue


class ResidueGraft:
    """
    A residue graft describes the change of a residue type and the coordinates of its atoms, based on datapoints from
    a pocket cluster.
    """

    def __init__(self, binder_res: DesignedBinderResidue, pocket_cluster: PocketCluster, penalty: float):
        """Creates a new instance.
        :param binder_res: the binder residue to be mutated from the acomplex
        :param pocket_cluster: the pocket cluster that provides type and atom coordinates of the grafted residue
        :param penalty: deviation of the cluster centroid from the binder residue's position and orientation
        """
        self.binder_res = binder_res
        self.pocket_cluster = pocket_cluster
        self.penalty = penalty

    def __repr__(self):
        return f"<ResidueGraft {self.binder_res.original_residue_type}{self.binder_res.residue_id}: " \
               f"{self.pocket_cluster.binder_restype}>"


class PocketGraft:
    """
    A pocket graft represents a set of binder residue grafts, which have been determined based on the properties of
    one given ligand residue.
    """

    def __init__(self, ligand_res: DesignedLigandResidue, residue_grafts: List[ResidueGraft]):
        """Creates a new instance.
        :param ligand_res: the ligand residue whose surrounding binder residues are subject to grafting
        :param residue_grafts: the residue grafts contained by this pocket graft.
        """
        self.ligand_res = ligand_res
        self.residue_grafts = residue_grafts

    def __repr__(self):
        return f"<PocketGraft {self.ligand_res.restype}{self.ligand_res.residue_id}: {self.residue_grafts}>"

    def penalty(self) -> float:
        """
        :return: Mean penalty of all residue grafts contained by this pocket graft
        """
        if len(self.residue_grafts) == 0:
            return 0.0
        return sum([rg.penalty for rg in self.residue_grafts]) / float(len(self.residue_grafts))


class BinderGraft:
    """
    A binder graft consists of several residue grafts, grouped by pocket grafts.
    """

    def __init__(self, pocket_grafts: List[PocketGraft]):
        """Creates a new instance
        :param pocket_grafts: disjoint partition of residue grafts.
        """
        self.pocket_grafts = pocket_grafts

    def __repr__(self):
        return f"<BinderGraft {self.pocket_grafts}>"

    def penalty(self) -> float:
        """
        :return: Mean penalty of all pocket grafts contained by this binder graft
        """
        if len(self.pocket_grafts) == 0:
            return 0.0
        return sum([pg.penalty() for pg in self.pocket_grafts]) / float(len(self.pocket_grafts))


def calculate_penalty(binder_res: DesignedBinderResidue, cluster_centroid: AtlasDatapoint,
                      distance_factor: float, orient_factor: float, secor_factor: float) -> float:
    """Calculates the distance between a wildcard residue and the centroid of a cluster.
    :param binder_res:
    :param cluster_centroid:
    :param distance_factor: weight factor for the distance term [1/Angsrom]
    :param orient_factor: weight factor for primary orientation [1/radians]
    :param secor_factor: weight factor for secondary orientation [1/radians]
    :return: a positive real value reflecting the penalty
    """
    calpha_distance: float = abs((binder_res.calpha() - cluster_centroid.binder_calpha()).norm())
    orient_angle: float = abs(binder_res.orientation().angle(cluster_centroid.binder_orientation()))
    sec_orient_angle: float = abs(binder_res.secondary_orientation().angle(
        cluster_centroid.binder_secondary_orientation()))
    return distance_factor * calpha_distance + orient_factor * orient_angle + secor_factor * sec_orient_angle


def get_optimal_pocket_graft(ligand_residue: DesignedLigandResidue, wildcard_residues: List[DesignedBinderResidue],
                             pocket: Pocket, penalty_threshold: float,
                             distance_factor: float, orient_factor: float, secor_factor: float) -> PocketGraft:
    """Yields the optimal assignment between wildcard residues and pocket members, such that the sum of penalties
    between the wildcard residue and the pocket's centroid is globally minimal. The returned assignment in terms of
    pocket grafts may leave wildcard residues unbound
    :param ligand_residue: the ligand residue where the pocket is to be grafted
    :param wildcard_residues: to be assigned to pocket members
    :param pocket: its members are assigned wildcard residues
    :param penalty_threshold: if the penalty exceeds this threshold, corresponding pocket grafts are ignored
    :param distance_factor: see calculate_penalty
    :param orient_factor: see calculate_penalty
    :param secor_factor: see calculate_penalty
    :return: a pocket graft that reflects the calculated assignment
    """
    penalties: Dict[Tuple[int, int], float] = {}
    for i, wcres in enumerate(wildcard_residues):
        for j, cluster in enumerate(pocket.clusters):
            pen = calculate_penalty(wcres, cluster.centroid, distance_factor, orient_factor, secor_factor)
            penalties[(i, j)] = pen
    occupied_is: Set[int] = set()
    occupied_js: Set[int] = set()
    residue_grafts: List[ResidueGraft] = []
    for (i, j), pen in sorted(list(penalties.items()), key=itemgetter(1)):
        if len(residue_grafts) == len(wildcard_residues) or pen > penalty_threshold:
            break
        if i not in occupied_is and j not in occupied_js:
            residue_graft = ResidueGraft(wildcard_residues[i], pocket.clusters[j], pen)
            residue_grafts.append(residue_graft)
            occupied_is.add(i)
            occupied_js.add(j)
    return PocketGraft(ligand_residue, residue_grafts)


def get_pocket_grafts_for_candidates(ligand_residue: DesignedLigandResidue,
                                     wildcard_residues: List[DesignedBinderResidue], pocket_candidates: List[Pocket],
                                     penalty_threshold: float, distance_factor: float, orient_factor: float,
                                     secor_factor: float) -> List[PocketGraft]:
    """Calculates the optimal pocket graft for every element in a list of candidate pockets. Orders the candidates by
    the overall penalty their grafting would imply (assuming that nothing has been grafted yet)
    :param ligand_residue: see get_optimal_pocket_graft
    :param wildcard_residues: see get_optimal_pocket_graft
    :param pocket_candidates: candidates whose optimal grafts are calculated and ordered by penalty
    :param penalty_threshold: see get_optimal_pocket_graft
    :param distance_factor: see calculate_penalty
    :param orient_factor: see calculate_penalty
    :param secor_factor: see calculate_penalty
    :return: a list of pocket grafts corresponding to the candidates passed, ordered by increasing penalty
    """
    return sorted([get_optimal_pocket_graft(ligand_residue, wildcard_residues, p, penalty_threshold, distance_factor,
                                            orient_factor, secor_factor) for p in pocket_candidates],
                  key=lambda pg: pg.penalty())


def get_optimal_binder_graft(pocket_grafts: Dict[int, List[PocketGraft]], max_binder_grafts: int = 1) \
        -> List[BinderGraft]:
    """ Given a dictionary of containing lists of pocket grafts, which potentially contain contradicting grafts
    for equivalent residues, this function selects the binder grafts with minimal penalty containing as many as possible
    pocket grafts (one per ligand position).
    :param pocket_grafts: pocket grafts obtained by local optimization of ligand pockets. In a dictionary with indices
    (range(x)) as keys (for the different ligand positions) and a list of pocket grafts as values (ordered by penalty
    - should not be required though).
    :param max_binder_grafts: Maximum number of binder grafts returned
    :return: a list of minimum-penalty and non-contradictory binder grafts, sorted by ascending penalty
    """

    def get_combinations(nones: int = 0):
        if not nones:
            return product(*(range(len(x)) for x in pocket_grafts.values()))

        none_combinations = combinations(range(n_pockets), nones)
        combin = []
        for none_positions in none_combinations:
            x_list = list(range(len(x)) for k, x in pocket_grafts.items() if k not in none_positions)
            for none in none_positions:
                x_list.insert(none, [None])
            combin.extend(list(product(*x_list)))
        return combin

    def get_penalty(graft_ids):
        return sum(penalty_dict[gi_n][gi] for gi_n, gi in enumerate(graft_ids))

    def mutually_exclusive(c):
        sets = [residue_sets[lg][pg] for lg, pg in enumerate(c)]
        return not any(set_c[0] & set_c[1] for set_c in combinations(sets, 2))

    residue_sets = {}
    penalty_dict = {}
    for lig, pgs in pocket_grafts.items():
        residue_sets[lig] = {}
        penalty_dict[lig] = {}
        for i, pg in enumerate(pgs):
            residue_sets[lig][i] = {(rg.binder_res.chain, rg.binder_res.residue_id) for rg in pg.residue_grafts}
            penalty_dict[lig][i] = pg.penalty()
        residue_sets[lig][None] = set()
        penalty_dict[lig][None] = 0.0

    n_pockets = len(list(pocket_grafts))
    binder_grafts = []
    n_lacking_pockets = 0
    while n_lacking_pockets < n_pockets:
        for comb in sorted(get_combinations(n_lacking_pockets), key=get_penalty):
            if mutually_exclusive(comb):
                binder_grafts.append(
                    BinderGraft(pocket_grafts=[pocket_grafts[ci][c] for ci, c in enumerate(comb) if c is not None])
                )
                if len(binder_grafts) >= max_binder_grafts:
                    return binder_grafts
        n_lacking_pockets += 1
    return binder_grafts


def select_best_pocket_graft(candidates: List[PocketGraft]) -> PocketGraft:
    """Default strategy for selecting an element from the list of pocket candidates (ordered by penalty). Take the
    cheapest element.
    :param candidates:
    :return: the first (i.e., lowest-penalty) candidate
    """
    return candidates[0]


def get_optimal_graft_for_acomplex(acomplex: LigandBinderComplex, pockets: Dict[str, List[Pocket]],
                                   penalty_threshold: float, distance_factor: float, orient_factor: float,
                                   secor_factor: float,
                                   select: Callable[[List[BinderGraft]], BinderGraft] = select_best_pocket_graft,
                                   loc_observer: Callable[[int, int], None] = None,
                                   glob_observer: Callable[[int, int], None] = None) -> BinderGraft or None:
    """ Based on a ligand/binder complex and a pocket dictionary, this method calculates an optimal binder graft based
    on a configurable candidate selection strategy
    :param acomplex: the ligand binder complex that defines both ligand residues and wildcards in the binder
    :param pockets: the output of a pocket mining run providing the basis for pocket selection
    :param penalty_threshold: see get_optimal_pocket_graft
    :param distance_factor: see calculate_penalty
    :param orient_factor: see calculate_penalty
    :param secor_factor: see calculate_penalty
    :param select: the selection strategy to apply for choosing a binder graft
    :param loc_observer: callback to react to the progress of the local optimization stage
    :param glob_observer: callback to react to the progress of the local optimization stage
    :return: a globally optimal binder graft, or None if selection was cancelled
    """
    pocket_grafts: Dict[int, List[PocketGraft]] = {}
    loc_len = len(acomplex.ligand.residues)
    if loc_observer is not None:
        loc_observer(0, loc_len)
    for i, ligres in enumerate(acomplex.ligand.residues):
        icoor = get_icoor_from_aresidue(ligres)
        wc_res = [br.transform_coor(icoor, True) for br in acomplex.binder.residues]
        ligpockets = pockets[ligres.restype]
        pgs = get_pocket_grafts_for_candidates(ligres, wc_res, ligpockets, penalty_threshold, distance_factor,
                                               orient_factor, secor_factor)
        pocket_grafts[i] = pgs
        if loc_observer is not None:
            loc_observer(i, loc_len)

    binder_grafts = get_optimal_binder_graft(pocket_grafts=pocket_grafts, max_binder_grafts=1)
    binder_graft = select(binder_grafts)
    if binder_graft is None:
        return None
    glob_len = len(binder_graft.pocket_grafts)
    if glob_observer is not None:
        glob_observer(0, glob_len)
    for i, pg in enumerate(binder_graft.pocket_grafts):
        icoor = get_icoor_from_aresidue(pg.ligand_res)
        for rg in pg.residue_grafts:
            binder_res: DesignedBinderResidue = rg.binder_res
            rg.binder_res = binder_res.transform_coor(icoor, False)
        if glob_observer is not None:
            glob_observer(i, glob_len)
    return binder_graft


def apply_residue_graft(binder_residue: Residue, residue_graft: ResidueGraft,
                        distance_factor: float, orient_factor: float, secor_factor: float) -> None:
    """Applies the modifications described by a residue graft to a PDB residue belonging to the binder
    :param binder_residue: the affected PDB binder residue
    :param residue_graft: describes the modification to apply
    :param distance_factor: see calculate_penalty
    :param orient_factor: see calculate_penalty
    :param secor_factor: see calculate_penalty
    """
    crep = residue_graft.pocket_cluster.get_most_representative_member(distance_factor, orient_factor, secor_factor)
    crep_icoor = get_icoor_from_aresidue(AbstractResidue(crep.binder_atoms))
    binres_icoor = get_icoor_from_aresidue(residue_graft.binder_res)
    binder_residue.resname = crep.binder_restype
    terminal = 'OXT' in binder_residue
    binder_residue.child_list = []
    binder_residue.child_dict = {}
    for atom_name, atom_position in crep.binder_atoms.items():
        if (crep.binder_restype == 'GLY' and atom_name == 'CB') or (not terminal and atom_name == 'OXT'):
            continue
        rel_position = crep_icoor.external_to_internal(atom_position, False)
        new_position = binres_icoor.internal_to_external(rel_position, False)
        atom = Atom(atom_name, new_position, 0.0, 1.0, ' ', atom_name, 0, atom_name[0])
        binder_residue.add(atom)


def apply_pocket_graft(ligand_residue: Residue, pocket_graft: PocketGraft,
                       distance_factor: float, orient_factor: float, secor_factor: float) -> None:
    """Updates the ligand coordinates based on the information obtained from a pocket graft.
    :param ligand_residue: the affected PDB ligand residue
    :param pocket_graft: this graft is made effective
    :param distance_factor: see calculate_penalty
    :param orient_factor: see calculate_penalty
    :param secor_factor: see calculate_penalty
    """
    if len(pocket_graft.residue_grafts) == 0:
        return
    crep = pocket_graft.residue_grafts[0].pocket_cluster.get_most_representative_member(distance_factor, orient_factor,
                                                                                        secor_factor)
    icoor = get_icoor_from_aresidue(pocket_graft.ligand_res)
    ligand_residue.resname = crep.ligand_restype
    terminal = 'OXT' in ligand_residue
    ligand_residue.child_list = []
    ligand_residue.child_dict = {}
    for atom_name, atom_position in crep.ligand_atoms.items():
        if (crep.ligand_restype == 'GLY' and atom_name == 'CB') or (not terminal and atom_name == 'OXT'):
            continue
        if atom_name in ligand_residue:
            ligand_residue.detach_child(atom_name)
        new_position = icoor.internal_to_external(atom_position, False)
        atom = Atom(atom_name, new_position, 0.0, 1.0, ' ', atom_name, 0, atom_name[0])
        ligand_residue.add(atom)


def apply_binder_graft(scaffold: Structure, binder_graft: BinderGraft, distance_factor: float, orient_factor: float,
                       secor_factor: float, res_lock: List[Tuple[int, str]] = None) -> None:
    """Applies all pocket grafts and residue grafts contained in this binder graft to a given scaffold structure
    :param scaffold: the scaffold PDB structure that will be modified in-place.
    :param binder_graft: the binder graft to apply
    :param distance_factor: see calculate_penalty
    :param orient_factor: see calculate_penalty
    :param secor_factor: see calculate_penalty
    :param res_lock: Locks binder residues and prevents them beeing mutated; List of Tuples containing residue id, type
    """

    def apply_residue_graft_in_presence_of_manual_mut():
        # If res_lock is defined, the scaffold defines some predefined mutations. Thus we have to check, if we would
        # destroy mutations with our residue graft. Plus mutations to same residue type is allowed, because predefined
        # mutations usually give no side chain information which we can add here.
        if (rg.binder_res.residue_id, newres) in res_lock or \
                rg.binder_res.residue_id not in (lock[0] for lock in res_lock):
            apply_residue_graft(res, rg, distance_factor, orient_factor, secor_factor)

    def apply_residue_graft_in_absence_of_manual_mut():
        # If res_lock is not defined, we did not define manual mutations to the scaffold. Thus, we don't have to check
        # the res_lock and we won't allow mutations to the original types.
        if newres != rg.binder_res.original_residue_type:
            print(f'{chain.get_id()}: {rg.binder_res.original_residue_type}{res.get_id()[1]}{newres}')
            apply_residue_graft(res, rg, distance_factor, orient_factor, secor_factor)

    if res_lock is None:
        apply_residue_graft_ = apply_residue_graft_in_absence_of_manual_mut
    else:
        apply_residue_graft_ = apply_residue_graft_in_presence_of_manual_mut

    for pg in binder_graft.pocket_grafts:
        for rg in pg.residue_grafts:
            for chain in scaffold.get_chains():
                if chain.get_id() == rg.binder_res.chain:
                    for res in chain.get_residues():
                        if res.get_id()[1] == rg.binder_res.residue_id:
                            newres = rg.pocket_cluster.get_most_representative_member(distance_factor, orient_factor,
                                                                                      secor_factor).binder_restype
                            apply_residue_graft_()
        for chain in scaffold.get_chains():
            if chain.get_id() == pg.ligand_res.chain:
                for res in chain.get_residues():
                    if res.get_id()[1] == pg.ligand_res.residue_id:
                        apply_pocket_graft(res, pg, distance_factor, orient_factor, secor_factor)


def simple_graft(scaffold: Structure, descriptor: ComplexDescriptor, pockets: Dict[str, List[Pocket]],
                 penalty_threshold: float, distance_factor: float, orient_factor: float, secor_factor: float,
                 select: Callable[[List[BinderGraft]], BinderGraft] = select_best_pocket_graft,
                 loc_observer: Callable[[int, int], None] = None,
                 glob_observer: Callable[[int, int], None] = None) -> Tuple[Union[BinderGraft, None], Structure]:
    """This function implements a simple algorithm for mutating the residues of a binder protein to match a ligand
    with given sequence. To find these mutations, information from the pocket miner is exploited.
    :param scaffold: the scaffold PDB structure that is modified in-place
    :param descriptor: describes the ligand residues as well as the wildcards of the binder chain
    :param pockets: result of the pocket miner where the algorithm extracts its information from
    :param penalty_threshold: see get_optimal_pocket_graft
    :param distance_factor: see calculate_penalty
    :param orient_factor: see calculate_penalty
    :param secor_factor: see calculate_penalty
    :param select: the selection strategy to apply for choosing a pocket graft
    :param loc_observer: callback to react to the progress of the local optimization stage
    :param glob_observer: callback to react to the progress of the local optimization stage
    :return: the binder graft representing the best solution
    """
    design = copy.deepcopy(scaffold)
    acomplex = generate_complex(None, design, descriptor)

    opt_graft = get_optimal_graft_for_acomplex(acomplex=acomplex, pockets=pockets,
                                               penalty_threshold=penalty_threshold, distance_factor=distance_factor,
                                               orient_factor=orient_factor, secor_factor=secor_factor,
                                               select=select, loc_observer=loc_observer, glob_observer=glob_observer)
    if opt_graft is not None:
        apply_binder_graft(design, opt_graft, distance_factor, orient_factor, secor_factor)
    return opt_graft, design


def multi_graft(scaffold: Structure, descriptor: ComplexDescriptor, pockets: Dict[str, List[Pocket]],
                penalty_threshold: float, distance_factor: float, orient_factor: float, secor_factor: float,
                n_solutions: int = 20, observer: Callable[[int], None] = None) \
        -> List[Tuple[float, BinderGraft, Structure]]:
    """A variant of the simple graft algorithm; the top n pocket grafts are globally optimized and the different
    corresponding binder grafts are returned
    :param scaffold: the scaffold PDB structure that is modified in-place
    :param descriptor: describes the ligand residues as well as the wildcards of the binder chain
    :param pockets: result of the pocket miner where the algorithm extracts its information from
    :param penalty_threshold: see get_optimal_pocket_graft
    :param distance_factor: see calculate_penalty
    :param orient_factor: see calculate_penalty
    :param secor_factor: see calculate_penalty
    :param n_solutions: maximum number of solutions to generate
    :param observer: callback to react to the index of the solution currently processed
    :return: a list of max size n_solutions, which contains the penalty, the binder graft, and the mutated structure
    belonging to every solution
    """

    def _select_ith_pocket_graft(candidates: List[BinderGraft]) -> BinderGraft:
        try:
            return candidates[i]
        except IndexError:
            pass

    result: List[Tuple[float, BinderGraft, Structure]] = []
    for i in range(n_solutions):
        binder_graft, structure = simple_graft(scaffold, descriptor, pockets, penalty_threshold, distance_factor,
                                               orient_factor, secor_factor, select=_select_ith_pocket_graft)
        if observer is not None:
            observer(i)
        if binder_graft is None:
            break
        result.append((binder_graft.penalty(), binder_graft, structure))
    return result


def get_mutations_single_p_graft(pocket: Pocket, scaffold: Structure, descriptor: ComplexDescriptor, res_id: int = 0,
                                 penalty_threshold: float = 32.0, distance_factor: float = 2.0,
                                 orient_factor: float = 1.0,
                                 secor_factor: float = 1.0, res_lock: List[Tuple[int, str]] = None) \
        -> Tuple[PocketGraft, Structure]:
    acomplex = generate_complex(None, scaffold, descriptor)
    pocket_grafts: List[PocketGraft] = []

    # Take ligand residue with correct residue id
    ligres = None
    for lres_i, lres in enumerate(acomplex.ligand.residues):
        if lres.residue_id == res_id:
            ligres = acomplex.ligand.residues[lres_i]
            break
    assert ligres is not None

    icoor = get_icoor_from_aresidue(ligres)
    wc_res = [br.transform_coor(icoor, True) for br in acomplex.binder.residues]
    pocket_graft = get_optimal_pocket_graft(ligres, wc_res, pocket, penalty_threshold, distance_factor,
                                            orient_factor, secor_factor)
    for i, rg in enumerate(pocket_graft.residue_grafts):
        pocket_graft.residue_grafts[i].binder_res = rg.binder_res.transform_coor(icoor, False)

    pocket_grafts.append(pocket_graft)
    binder_graft = BinderGraft(pocket_grafts)  # No need to find best graft, because we only have one
    apply_binder_graft(scaffold, binder_graft, distance_factor, orient_factor, secor_factor, res_lock=res_lock)
    return pocket_graft, scaffold


def prepare_scaffold_and_descriptor_for_grafting(scaffold_name: str,
                                                 ligand_residues: Union[Tuple[str, str],
                                                                        Dict[str, Dict[Union[str, int], str]]],
                                                 design_positions: Dict[str, List[Union[str, int]]],
                                                 pocket_restype: str = None):
    """
    Prepares the scaffold structure (Bio.PDB) from path string and the ComplexDescriptor from definitions
    of ligand residues and binder residues for subsequent grafting.

    :param scaffold_name: the path to a pdb structure including the chain ids and res ids given in ligand_residue
    and design_positions.
    :param ligand_residues:
    Either: A tuple of chain id, residue id (str or int) and desired residue type (needs a pocket_restype definition!)
    OR: A dict of chain id matched with a dict of residue id (str or int) and desired residue type.
    :param design_positions: A dictionary with the keys being the chain ids and the values being a list of all residue
    ids that should be allowed to be mutated while grafting.
    :param pocket_restype: The residue type the ligand should be mutated to. Only used if ligand_residues is a tuple.
    :return: scaffold as Bio.PDB structure and ComplexDescriptor
    """

    scaffold = PDBParser(QUIET=True).get_structure(scaffold_name,
                                                   scaffold_name)
    if isinstance(ligand_residues, dict):
        ligand_info = []
        for ligand_chain, ligand_chain_residues in ligand_residues.items():
            for ligand_res_id, ligand_restype in ligand_chain_residues.items():
                ligand_info.append((ligand_chain, str(ligand_res_id), ligand_restype))
    else:
        ligand_chain, ligand_res_id = ligand_residues
        ligand_info = [(ligand_chain, str(ligand_res_id), pocket_restype)]
        assert pocket_restype is not None
    binder_info = []
    for chain, positions in design_positions.items():
        binder_info += list((chain, str(pos)) for pos in positions)
    coupling_info = []
    descriptor = ComplexDescriptor(ligand_info, binder_info, coupling_info)

    return scaffold, descriptor


def graft_pocket_onto_scaffold(pocket: Pocket, scaffold_name: str, ligand_residue: Tuple[str, Union[str, int]],
                               design_positions: Dict[str, List[Union[str, int]]], output_name=None) -> Structure:
    """
    Will graft the residues of pocket onto the pdb structure found at scaffold_name.
    The grafting is applied based on the ligand residue position and the design_positions (where mutations are
    permitted).

    :param pocket: a Pocket object containing at least one member per cluster
    :param scaffold_name: the path to a pdb structure including the chain ids and res ids given in ligand_residue
    and design_positions.
    :param ligand_residue: A tuple of chain id, residue id (str or int) and desired residue type.
    :param design_positions: A dictionary with the keys being the chain ids and the values being a list of all residue
    ids that should be allowed to be mutated while grafting.
    :param output_name: If given this will be the path of the newly designed pdb structure
    :return: The newly designed Bio.PDB structure
    """
    scaffold, descriptor = prepare_scaffold_and_descriptor_for_grafting(scaffold_name=scaffold_name,
                                                                        ligand_residues=ligand_residue,
                                                                        design_positions=design_positions,
                                                                        pocket_restype=pocket.restype)

    graft, design = get_mutations_single_p_graft(pocket, scaffold, descriptor, res_id=int(ligand_residue[1]))
    if output_name is not None:
        io = PDBIO()
        io.set_structure(design)
        io.save(output_name)

    return design


def graft_best_pocket_onto_scaffold(pockets: Dict[str, List[Pocket]], scaffold_name: str,
                                    ligand_residues: Dict[str, Dict[Union[str, int], str]],
                                    design_positions: Dict[str, List[Union[str, int]]], output_name=None,
                                    penalty_threshold: float = 32.0, distance_factor: float = 2.0,
                                    orient_factor: float = 1.0, secor_factor: float = 1.0, ):
    """
    Will graft the residues of the best matching pocket in pockets onto the pdb structure found at scaffold_name.
    The grafting is applied based on the ligand residue positions and the design_positions (where mutations are
    permitted) and based on the defined ligand residue types.

    :param pockets: a collection of Pockets (result of the pocket miner) containing at least one member per cluster
    and at least one pocket where the restype is the ligand_restype.
    :param scaffold_name: the path to a pdb structure including the the chain ids and res ids given in ligand_residue
    and design_positions.
    :param ligand_residues: A dict of chain id matched with a dict of residue id (str or int) and desired residue type.
    :param design_positions: A dictionary with the keys being the chain ids and the values being a list of all residue
    ids that should be allowed to be mutated while grafting.
    :param output_name: If given this will be the path of the newly designed pdb structure
    :param penalty_threshold: see get_optimal_pocket_graft
    :param distance_factor: see calculate_penalty
    :param orient_factor: see calculate_penalty
    :param secor_factor: see calculate_penalty
    :return: A tuple containing the newly designed Bio.PDB structure and the BinderGraft applied
    """
    scaffold, descriptor = prepare_scaffold_and_descriptor_for_grafting(scaffold_name=scaffold_name,
                                                                        ligand_residues=ligand_residues,
                                                                        design_positions=design_positions)

    graft, design = simple_graft(scaffold=scaffold, descriptor=descriptor, pockets=pockets,
                                 penalty_threshold=penalty_threshold, distance_factor=distance_factor,
                                 orient_factor=orient_factor, secor_factor=secor_factor)
    if output_name is not None:
        io = PDBIO()
        io.set_structure(design)
        io.save(output_name)

    return design, graft


def multi_graft_best_pocket_onto_scaffold(pockets: Dict[str, List[Pocket]], scaffold_name: str,
                                          ligand_residues: Dict[str, Dict[Union[str, int], str]],
                                          design_positions: Dict[str, List[Union[str, int]]], output_name=None,
                                          penalty_threshold: float = 32.0, distance_factor: float = 2.0,
                                          orient_factor: float = 1.0, secor_factor: float = 1.0,
                                          n_solutions: int = 10):
    """
    Does the same thing as graft_best_pocket_onto_scaffold, but returns the best n_solutions solutions of grafting:

    Will graft the residues of the best matching pocket in pockets onto the pdb structure found at scaffold_name.
    The grafting is applied based on the ligand residue positions and the design_positions (where mutations are
    permitted) and based on the defined ligand residue types.

    :param pockets: a collection of Pockets (result of the pocket miner) containing at least one member per cluster
    and at least one pocket where the restype is the ligand_restype.
    :param scaffold_name: the path to a pdb structure including the the chain ids and res ids given in ligand_residue
    and design_positions.
    :param ligand_residues: A dict of chain id matched with a dict of residue id (str or int) and desired residue type.
    :param design_positions: A dictionary with the keys being the chain ids and the values being a list of all residue
    ids that should be allowed to be mutated while grafting.
    :param output_name: If given this will be the path of the newly designed pdb structures with added numbers for
    the n_th solution.
    :param penalty_threshold: see get_optimal_pocket_graft
    :param distance_factor: see calculate_penalty
    :param orient_factor: see calculate_penalty
    :param secor_factor: see calculate_penalty
    :param n_solutions: maximum number of solutions to generate
    :return: A list containing: penalty, BinderGraft and designed Bio.PDB structures (sorted best to worst).
    """
    scaffold, descriptor = prepare_scaffold_and_descriptor_for_grafting(scaffold_name=scaffold_name,
                                                                        ligand_residues=ligand_residues,
                                                                        design_positions=design_positions)

    result = multi_graft(scaffold=scaffold, descriptor=descriptor, pockets=pockets,
                         penalty_threshold=penalty_threshold, distance_factor=distance_factor,
                         orient_factor=orient_factor, secor_factor=secor_factor, n_solutions=n_solutions)

    if output_name is not None:
        if not output_name.endswith(".pdb"):
            output_name = output_name + ".pdb"
        for number, result_n in enumerate(result):
            design = result_n[2]
            io = PDBIO()
            io.set_structure(design)
            io.save(output_name.replace(".pdb", str(number + 1) + ".pdb"))

    return result



"""This package contains reusable components of ATLIGATOR -- the ATlas based LIGAnd site ediTOR.

:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:date: 2018-03-05
"""

__all__ = ["acomplex", "atlas", "base_scorer", "genetic_mutator", "geometry", "io_util", "mutation", "pdb_util",
           "pocket_miner", "prediction_matrix", "rosetta_util", "structure", "visualization"]



from typing import List

from Bio.PDB import PDBParser, Structure
from Bio.SeqUtils import seq1


def get_aa_from_resnum(structure: str, target: List[int] or int) -> List[str] or str:
    """
    Returns the residue name (one letter code) of a target residue (id or list of ids).
    The name is read out of the structure stored in structure parameter path.
    :param structure: path of structure to use as search space
    :param target: int or list of ints to check for residue name
    :return: residue name or list of residue names corr. to target
    """
    parser = PDBParser(QUIET=True)
    mol: Structure = parser.get_structure(structure, structure)
    if type(target) == int:
        for residue in mol.get_residues():
            if residue.get_id()[1] == target:
                return seq1(residue.get_resname())
        return None
    else:
        resnames = []
        for res_id in target:
            for residue in mol.get_residues():
                if residue.get_id()[1] == res_id:
                    resnames.append(seq1(residue.get_resname()))
    return resnames



"""This module provides an implementation of the BaseScorer interface that utilizes the Osprey protein design suite.
Unfortunately, the Osprey python interface seems to be strictly limited to python2. Therefore, we use another
indirection here: our python3 module creates a python2 script, which is executed and starts the Osprey Java app.

For installing Osprey we used a local installation ('python2 get-pip.py --user' from 'https://bootstrap.pypa.io/get-pip.py') of pip (pip2! NOT inside conda env!).
After downloading (https://github.com/donaldlab/OSPREY_refactor/releases) and unzipping the osprey package you have
to run following commands:
    'pip2 install osprey --no-index --user --find-link=wheelhouse'
    'pip2 install -U numpy --user'


THIS recommended version will NOT work: (wheel is deprecated and installation has to be done locally)
    'pip2 install osprey --no-index --use-wheel --find-link=wheelhouse'


:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:date: 2018-04-19
"""

from subprocess import run

from atligator.acomplex import read_complex_descriptor
from atligator.base_scorer import BaseScorer, ScoreResult
from atligator.genetic_mutator import ComplexDescriptor
from atligator.pdb_util import replace_restypes


class OspreyScorer(BaseScorer):
    """Implements the BaseScorer interface upon Osprey."""

    def __init__(self, sidechain_flexibility: str, astar_mode: str, finder: str, use_epic: bool, use_lute: bool,
                 python2_executable: str = "python2", descriptor: ComplexDescriptor = None, cores: int = None):
        """Creates a new instance.
        :param sidechain_flexibility: Resolution for sidechain optimization. Must be one of ['library', 'discrete',
        'continuous']
        :param astar_mode: A* implementation to use. Must be one of ['traditional', 'mplp]
        :param finder: Finder method to use for optimization. Must be one of ['gmec', 'deegmec']
        :param use_epic: whether to use the EPIC algorithm for continuous flexibility modeling
        :param use_lute: whether to use the LUTE algorithm
        :param python2_executable: the path to python2 to pass to the shell
        :param descriptor: path to descriptor file to read out mutable binder and ligand residues
        :param cores: number of cores used for energy calculation
        """
        if sidechain_flexibility not in ['library', 'discrete', 'continuous', 'wildtype']:  # 'wildtype' for rigid
            raise ValueError()
        self.sidechain_flexibility = sidechain_flexibility
        if astar_mode not in ['traditional', 'mplp']:
            raise ValueError()
        self.astar_mode = astar_mode
        if finder not in ['gmec', 'deegmec', 'lute']:  # 'lute' for new lute_train method and testing
            raise ValueError()
        self.finder = finder
        self.use_epic = use_epic
        self.use_lute = use_lute
        self.python2_executable = python2_executable
        if descriptor is None:
            self.descriptor = read_complex_descriptor("./descriptor")
        else:
            self.descriptor = descriptor
        self.cores = 1 if cores is None else cores

    def score(self, structure: str, descriptor: ComplexDescriptor) -> ScoreResult:
        """
        Optimizes and scores the structure according to the Osprey parameters set in the constructor.
        :param structure: structure to optimize and score
        :param descriptor: defines ligand and binder residues
        :return: optimized structure and score obtained
        """
        replace_restypes(structure, 'HIS', 'HID')
        optimized_structure = structure.replace('.pdb', '_opt.pdb')
        res_bin = []
        res_lig = []
        for i in self.descriptor.binder_info:
            res_bin.append(i[0] + i[1])
        for i in self.descriptor.ligand_info:
            res_lig.append(i[0] + i[1])

        print(res_lig, res_bin)
        # Last few lines are needed to read out which residues to set flexible (mutatable ones from descriptor)
        # To toggle between these and all residues change commenting in the lines:
        # f"for res in strand.mol.residues:\n" + \
        # f"#for resname in {res_bin}:\n" + \
        # f"    resname = res.pDBResNumber\n"
        py2script = f"import osprey\n" + \
                    f"osprey.start()\n" + \
                    f"strand = osprey.Strand('{structure}')\n" + \
                    f"#for res in strand.mol.residues:\n" + \
                    f"for resname in {res_bin+res_lig}:\n" + \
                    f"#    resname = res.pDBResNumber\n"
        # .setLibraryRotamers(osprey.WILD_TYPE) sets to any residue type specified in the brackets. In this case
        # a library of rotamers of the wild type residue type (aa) is used. empty brackets would give same result.
        # .addWildTypeRotamers() adds the wild type rotamer. So with 'wildtype' only the original rotamer (or rotamers
        # if there are several conformations) is used.
        if self.sidechain_flexibility == 'library':
            py2script += f"    strand.flexibility[resname].setLibraryRotamers(osprey.WILD_TYPE).addWildTypeRotamers()\n"
        elif self.sidechain_flexibility == 'discrete':
            py2script += f"    strand.flexibility[resname].setLibraryRotamers(osprey.WILD_TYPE).setDiscrete()\n"
        elif self.sidechain_flexibility == 'wildtype':
            py2script += f"    strand.flexibility[resname].addWildTypeRotamers()\n"
        else:
            py2script += f"    strand.flexibility[resname].setLibraryRotamers(osprey.WILD_TYPE).setContinuous()\n"
        # Not sure if ff_params.solvScale = 0. is necessary. Found in one of x examples.
        #            f"bbflex = osprey.DEEPerStrandFlex(strand, '1CC8.pert', {res_lig}, '{structure}')\n" \
        #            f"conf_space = osprey.ConfSpace([[strand, bbflex]])\n" + \
        py2script += f"conf_space = osprey.ConfSpace(strand)\n" + \
                     f"ff_params = osprey.ForcefieldParams()\n" + \
                     f"ff_params.solvScale = 0.\n" + \
                     f"ene_calc = osprey.EnergyCalculator(conf_space, ff_params, " \
                     f"parallelism=osprey.Parallelism(cpuCores={self.cores}))\n" + \
                     f"conf_calc = osprey.ConfEnergyCalculator(conf_space, ene_calc)\n" + \
                     f"ene_mat = osprey.EnergyMatrix(conf_calc)\n"
        if self.astar_mode == 'mplp':
            py2script += f"astar_alg = osprey.AStarMPLP(ene_mat, conf_space)\n"
        else:
            py2script += f"astar_alg = osprey.AStarTraditional(ene_mat, conf_space)\n"
        if self.finder == 'gmec':
            py2script += f"finder = osprey.GMECFinder(astar_alg, conf_calc)\n" + \
                         f"solution = finder.find()\n"
        else:
            py2script += f"finder = osprey.DEEGMECFinder(ene_mat, conf_space, ene_calc, conf_calc, 'deegmec', " + \
                         f"use_epic={self.use_epic}, use_lute={self.use_lute})\n" + \
                         f"solution = finder.calcGMEC()\n"
        py2script += f"score = solution.getScore()\n" + \
                     f"optimized_molecule = conf_space.makeMolecule(solution.getAssignments())\n" + \
                     f"osprey.writePdb(optimized_molecule, '{optimized_structure}')\n" + \
                     f"print score\n" + \
                     f"print '{optimized_structure}'\n"
        # Step to use new lute_gmecFinder method of osprey. Not implemented in current version (3.0b1), but found in
        # examples/documentation and also found in git repository. If new osprey version gets released, uncomment this.
        # 71+72+75+78+79+81+82+83+88+113+114+117+120+121+123+124+125+130+155+156+159+162+163+165+166+167+172+197+198+201+204+205+207
        """
        if self.finder == 'lute':
            py2script = f"import osprey\n" + \
                        f"osprey.start()\n" + \
                        f"strand = osprey.Strand('{structure}')\n" \
                        f"for res in strand.mol.residues:\n" + \
                        f"    resname = res.pDBResNumber\n" \
                        f"    strand.flexibility[resname].setLibraryRotamers(osprey.WILD_TYPE).addWildTypeRotamers()\n" \
                        f"confSpace = osprey.ConfSpace(strand)\n" \
                        f"pmat = osprey.DEE_read(confSpace, 'LUTE.pmat.dat')\n" \
                        f"model = osprey.LUTE_read('LUTE.dat')\n" \
                        f"gmecFinder = osprey.LUTE_GMECFinder(confSpace, model, pmat, printIntermediateConfs=True)\n" \
                        f"gmec = gmecFinder.find(1.0)\n" \
                        f"score = solution.getScore()\n" \
                        f"optimized_molecule = conf_space.makeMolecule(solution.getAssignments())\n" + \
                        f"osprey.writePdb(optimized_molecule, '{optimized_structure}')\n" + \
                        f"print score\n" + \
                        f"print '{optimized_structure}'\n"
        """
        py2file = structure.replace(".pdb", "-osprey.py")
        with open(py2file, 'w') as fo:
            fo.write(py2script)
        py2out = structure.replace(".pdb", "-osprey.out")
        with open(py2out, 'w') as pfo:
            run(f"{self.python2_executable} {py2file}", shell=True, check=True, stdout=pfo)
        with open(py2out, 'r') as pfo:
            lines = list(pfo.readlines())
            score = float(lines[-2])
            optimized_structure = lines[-1].strip()
        replace_restypes(structure, 'HID', 'HIS')
        replace_restypes(optimized_structure, 'HID', 'HIS')
        return ScoreResult(score, optimized_structure)



"""Utility classes, functions, and data for the visualization of atlas data using matplotlib.

:Author: Josef Kynast <josef.kynast@uni-bayreuth.de>
:Author: Felix Schwaegerl <felix.schwaegerl@uni-bayreuth.de>
:date: 2018-03-05
"""
from collections import defaultdict
from itertools import combinations
from typing import Dict, List, Union, Tuple, Iterator

import matplotlib.pyplot as plt
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import plotly.io as pio
from Bio.PDB.vectors import Vector
# noinspection PyUnresolvedReferences
from mpl_toolkits.mplot3d import axes3d, Axes3D  # do not remove this import!

from atligator.acomplex import LigandBinderComplex
from atligator.atlas import Atlas, AtlasDatapoint
from atligator.pdb_util import canonical_amino_acids
from atligator.pocket_miner import Pocket
from atligator.prediction_matrix import proximity_score, BinderPredictionMatrix
from atligator.structure import get_icoor_from_aresidue, DesignedLigandResidue

pio.renderers.default = "browser"

# coloring scheme for protein residues, see doc/color-scheme.html
color_map: Dict[str, str] = {'ALA': '#82B2B8', 'ARG': '#035A12', 'ASN': '#A24689',
                             'ASP': '#DA675E', 'CYS': '#F7E748', 'GLN': '#570340',
                             'GLU': '#750C04', 'GLY': '#A6B9C1', 'HIS': '#FFD7B2',
                             'ILE': '#043F48', 'LEU': '#1B5F68', 'LYS': '#4A8658',
                             'MET': '#DDCF3E', 'PHE': '#DA995E', 'PRO': '#74D73E',
                             'SER': '#5B5BB3', 'THR': '#343477', 'TRP': '#753A04',
                             'TYR': '#AA672A', 'VAL': '#3B7C85'}

# coloring scheme for protein residues, see doc/color-scheme.html
color_map_light: Dict[str, str] = {'ALA': '#94cad1', 'ARG': '#047317', 'ASN': '#bd52a0',
                                   'ASP': '#F27268', 'CYS': '#FFF05C', 'GLN': '#700453',
                                   'GLU': '#8F0F05', 'GLY': '#BDD2DB', 'HIS': '#FFE5CC',
                                   'ILE': '#055561', 'LEU': '#227782', 'LYS': '#59A16A',
                                   'MET': '#F7E845', 'PHE': '#DA995E', 'PRO': '#F2AA68',
                                   'SER': '#6868CC', 'THR': '#404091', 'TRP': '#8F4705',
                                   'TYR': '#C47731', 'VAL': '#46939E'}

color_map_rasmol: Dict[str, str] = {'ALA': '#C8C8C8', 'ARG': '#145AFF', 'ASN': '#00DCDC',
                                    'ASP': '#E60A0A', 'CYS': '#E6E600', 'GLN': '#00DCDC',
                                    'GLU': '#E60A0A', 'GLY': '#EBEBEB', 'HIS': '#8282D2',
                                    'ILE': '#0F820F', 'LEU': '#0F820F', 'LYS': '#145AFF',
                                    'MET': '#E6E600', 'PHE': '#3232AA', 'PRO': '#DC9682',
                                    'SER': '#FA9600', 'THR': '#FA9600', 'TRP': '#B45AB4',
                                    'TYR': '#3232AA', 'VAL': '#0F820F'}

color_map_rasmol_light: Dict[str, str] = {'ALA': '#D7D7D7', 'ARG': '#3070FF', 'ASN': '#00FAF7', 'ASP': '#F41919',
                                          'CYS': '#FFFC04', 'GLN': '#00FAF7', 'GLU': '#F41919', 'GLY': '#FAFAFA',
                                          'HIS': '#9898D9', 'ILE': '#129B12', 'LEU': '#129B12', 'LYS': '#3070FF',
                                          'MET': '#FFFC04', 'PHE': '#3838BF', 'PRO': '#E2A899', 'SER': '#FFA019',
                                          'THR': '#FFA019', 'TRP': '#BC6EBD', 'TYR': '#3838BF', 'VAL': '#129B12'}

color_map_shapely: Dict[str, str] = {'ALA': '#8CFF8C', 'ARG': '#00007C', 'ASN': '#FF7C70',
                                     'ASP': '#A00042', 'CYS': '#FFFF70', 'GLN': '#FF4C4C',
                                     'GLU': '#660000', 'GLY': '#DFDFDF', 'HIS': '#7070FF',
                                     'ILE': '#004C00', 'LEU': '#455E45', 'LYS': '#4747B8',
                                     'MET': '#B8A042', 'PHE': '#534C42', 'PRO': '#525252',
                                     'SER': '#FF7042', 'THR': '#B84C00', 'TRP': '#4F4600',
                                     'TYR': '#8C704C', 'VAL': '#FF8CFF'}

color_map_shapely_light: Dict[str, str] = {'ALA': '#A9FFA9', 'ARG': '#00009A', 'ASN': '#FE958D', 'ASP': '#BE0050',
                                           'CYS': '#FEFD8D', 'GLN': '#FF6969', 'GLU': '#840000', 'GLY': '#EDEDED',
                                           'HIS': '#8D8DFE', 'ILE': '#006A00', 'LEU': '#516E51', 'LYS': '#5B5BC0',
                                           'MET': '#C2A955', 'PHE': '#635A4E', 'PRO': '#616161', 'SER': '#FF845F',
                                           'THR': '#D65500', 'TRP': '#6C5E00', 'TYR': '#9F7D56', 'VAL': '#FDA9FF'}

connections = {'ALA': ('CA', 'CB'),
               'ARG': ('CA', 'CB', 'CG', 'CD', 'NE', 'CZ', 'NH1', 'CZ', 'NH2'),
               'ASN': ('CA', 'CB', 'CG', 'OD1', 'CG', 'ND2'),
               'ASP': ('CA', 'CB', 'CG', 'OD1', 'CG', 'OD2'),
               'CYS': ('CA', 'CB', 'SG'),
               'GLN': ('CA', 'CB', 'CG', 'CD', 'OE1', 'CD', 'NE2'),
               'GLU': ('CA', 'CB', 'CG', 'CD', 'OE1', 'CD', 'OE2'),
               'GLY': (),
               'HIS': ('CA', 'CB', 'CG', 'CD2', 'NE2', 'CE1', 'ND1', 'CG'),
               'ILE': ('CA', 'CB', 'CG2', 'CB', 'CG1', 'CD1'),
               'LEU': ('CA', 'CB', 'CG', 'CD1', 'CG', 'CD2'),
               'LYS': ('CA', 'CB', 'CG', 'CD', 'CE', 'NZ'),
               'MET': ('CA', 'CB', 'CG', 'SD', 'CE'),
               'PHE': ('CA', 'CB', 'CG', 'CD2', 'CE2', 'CZ', 'CE1', 'CD1', 'CG'),
               'PRO': ('N', 'CD', 'CG', 'CB', 'CA'),
               'SER': ('CA', 'CB', 'OG'),
               'THR': ('CA', 'CB', 'OG1', 'CB', 'CG2'),
               'TRP': ('CA', 'CB', 'CG', 'CD1', 'NE1', 'CE2', 'CZ2', 'CH2', 'CZ3', 'CE3', 'CD2', 'CE2', 'CD2', 'CG'),
               'TYR': ('CA', 'CB', 'CG', 'CD1', 'CE1', 'CZ', 'OH', 'CZ', 'CE2', 'CD2', 'CG'),
               'VAL': ('CA', 'CB', 'CG2', 'CB', 'CG1')}

cmaps = {"0": color_map,
         "1": color_map_rasmol,
         "2": color_map_shapely}

cmaps_light = {"0": color_map_light,
               "1": color_map_rasmol_light,
               "2": color_map_shapely_light}


def get_color_map(color_mode: str = "0"):
    """
    :param color_mode: If specified this changes the color mapper to a different color scheme.
    :return: The associated color as hexcode, or #808080 if no corresponding entry found in the color map
    """
    try:
        return cmaps[color_mode]
    except KeyError:
        return cmaps["0"]


def get_color(restype: str, color_mode: str = "0") -> str:
    """Using color_map, this function looks up a color for the given amino acid residue.
    :param restype: the residue type in three-letter code to look up
    :param color_mode: If specified this changes the color mapper to a different color scheme.
    :return: The associated color as hexcode, or #808080 if no corresponding entry found in the color map
    """
    _color_map = get_color_map(color_mode)
    try:
        return _color_map[restype.upper()]
    except KeyError:
        return "#808080"


def get_light_color(restype: str, color_mode: str = "0") -> str:
    """Using color_map, this function looks up a color for the given amino acid residue.
    :param restype: the residue type in three-letter code to look up
    :param color_mode: If specified this changes the color mapper to a different color scheme.
    :return: The associated color as hexcode, or #808080 if no corresponding entry found in the color map
    """
    try:
        _color_map_light = cmaps_light[color_mode]
    except KeyError:
        _color_map_light = cmaps_light["0"]
    try:
        return _color_map_light[restype.upper()]
    except KeyError:
        return "#808080"


def visualize_atlas(atlas: Atlas, method: str = "plotly", **kwargs) -> None:
    """Shows the atlas datapoints with a plotly plot (within web broswer) or matplotlib (separate window).
    For additional parameters such as 'ligand_restype' compare the subfunctions used here.

    :param atlas: the atlas to visualize
    :param method: Either "plotly" or "matplotlib". If not plotly matplotlib is used.
    """
    if method == "plotly":
        return visualize_atlas_plotly(atlas=atlas, **kwargs)
    else:
        return visualize_atlas_matplotlib(atlas=atlas, **kwargs)


def visualize_atlas_plotly(atlas: Atlas = None, ligand_restype: Union[str, None] = None,
                           binder_restype: Union[str, None] = None, dimensions: float = 10.0, **_) -> None:
    """Shows a 3d plotly plot with in the web broswer that contains both the specified residue ligand
    and the atlas datapoints associated with the corresponding residue types. All residues are shown as
    Ca Cb markers with the Cb being smaller than Ca.

    :param atlas: the atlas to visualize
    :param ligand_restype: binder residues interacting with this ligand residue types will be displayed. If unset, no
    filtering is applied.
    :param binder_restype: binder residues of this type will be displayed. If unset, no filtering is applied.
    :param dimensions: max negative and positive coordinates to display, in Angstrom
    :param _: to prevent failure if matplotlib parameters are given.
    :return: None
    """

    if atlas is None:
        import pickle
        with open("/home/josef/Documents/alarms_test/L_a.118.1_.55a_.1b.atlas", "rb") as rf:
            atlas = pickle.load(rf)

    # Filter atlas, if filtering for ligand or binder residue is desired.
    if ligand_restype is not None or binder_restype is not None:
        atlas = atlas.filter(ligand_restype=ligand_restype, binder_restype=binder_restype)
    if ligand_restype is None:
        ligand_restype = "LIG"

    # collect all coordinates, residue types and the size of the marker datapoints in lists.
    x = []
    y = []
    z = []
    restypes = []
    size = []

    # Two datapoints representing ligand Calpha (x,y,z = 0,0,0) and Cbeta (1,0,0) are placed in the center.
    x.append(0.0)
    y.append(0.0)
    z.append(0.0)
    restypes.append("ligand " + ligand_restype)
    size.append(80)
    x.append(1.0)
    y.append(0.0)
    z.append(0.0)
    restypes.append("ligand " + ligand_restype + " cb")
    size.append(25)

    # For every datapoint save coords, restype and size (Ca and Cb)
    for dp in atlas.datapoints:
        x.append(dp.binder_calpha()[0])
        y.append(dp.binder_calpha()[1])
        z.append(dp.binder_calpha()[2])
        restypes.append(dp.binder_restype)
        size.append(20)

        x.append(dp.binder_cbeta()[0])
        y.append(dp.binder_cbeta()[1])
        z.append(dp.binder_cbeta()[2])
        restypes.append(dp.binder_restype + " cb")
        size.append(7)

    df = pd.DataFrame.from_dict({
        "x": x,
        "y": y,
        "z": z,
        "restype": restypes,
        "size": size,
    })

    # concatenate color dictionaries to address Ca (full color) and Cb (light color)
    color_map_light_cb = {k + " cb": v for k, v in color_map_light.items()}
    color_map_full = {**color_map, **color_map_light_cb}
    color_map_full["ligand " + ligand_restype] = color_map_full[ligand_restype] \
        if ligand_restype != "LIG" else "#808080"
    color_map_full["ligand " + ligand_restype + " cb"] = color_map_full[ligand_restype + " cb"] \
        if ligand_restype != "LIG" else "#808080"

    fig = px.scatter_3d(df, x='x', y='y', z='z',
                        color='restype', color_discrete_map=color_map_full, size="size")
    # size also introduces opacity and borders...

    # removing border, changing opacity.
    fig.update_traces(marker=dict(opacity=1.0,
                                  line=dict(width=0)),
                      selector=dict(name="ligand " + ligand_restype + " cb"))
    fig.update_traces(marker=dict(opacity=1.0,
                                  line=dict(width=0)),
                      selector=dict(name=ligand_restype + " cb"))
    for aa in canonical_amino_acids:
        fig.update_traces(marker=dict(opacity=1.0,
                                      line=dict(width=0)),
                          selector=dict(name=aa + " cb"))
        fig.update_traces(marker=dict(opacity=1.0,
                                      line=dict(width=0)),
                          selector=dict(name=aa))

    fig.update_layout(
        scene=dict(
            xaxis=dict(nticks=10, range=[-dimensions, dimensions], ),
            yaxis=dict(nticks=10, range=[-dimensions, dimensions], ),
            zaxis=dict(nticks=10, range=[-dimensions, dimensions], ),
            aspectmode="cube", ))

    fig.show()


def visualize_atlas_matplotlib(atlas: Atlas, ligand_restype: Union[str, None] = None,
                               binder_restype: Union[str, None] = None, title: str = None, dimensions: float = 10.0,
                               draw_secor: bool = False) -> None:
    """Shows a 3d quiver plot that contains both the specified residue ligand (including backbone) and the atlas
    datapoints associated with the corresponding ligand residue types.
    :param atlas: the atlas to visualize
    :param ligand_restype: binder residues interacting with this ligand residue types will be displayed. If unset, no
    filtering is applied.
    :param binder_restype: binder residues of this type will be displayed. If unset, no filtering is applied.
    :param title: this will be shown as the title of the plot
    :param dimensions: max negative and positive coordinates to display, in Angstrom
    :param draw_secor: if True, secondary orientation vectors are drawn (without arrowheads)
    """
    if title is None:
        if ligand_restype is None:
            title = 'Atlas'
        else:
            title = ligand_restype

    # Filter atlas, if filtering for ligand or binder residue is desired.
    if ligand_restype is not None or binder_restype is not None:
        atlas = atlas.filter(ligand_restype=ligand_restype, binder_restype=binder_restype)
    if ligand_restype is None:
        ligand_restype = "LIG"
    # plot the data as matplotlib 3d quiver
    fig = plt.figure()
    fig.suptitle(title)
    plt.suptitle(title)
    ax: Axes3D = fig.gca(projection='3d')
    ax.set_xlim(-dimensions, dimensions)
    ax.set_ylim(-dimensions, dimensions)
    ax.set_zlim(-dimensions, dimensions)
    # draw ligand residue
    ax.quiver([-1.0], [0.0], [0.0], [3.0], [0.0], [0.0],
              color=get_color(ligand_restype), arrow_length_ratio=0.4, linewidth=8)
    # process binder residues
    for binrestype in color_map:
        # for each potential interacting binder residue, paint the quivers in individual colors.
        relevant_datapoints = [d for d in atlas.datapoints if d.binder_restype == binrestype]
        col = color_map[binrestype]
        x: List[float] = []
        y: List[float] = []
        z: List[float] = []
        u: List[float] = []
        v: List[float] = []
        w: List[float] = []
        u2: List[float] = []
        v2: List[float] = []
        w2: List[float] = []
        for d in relevant_datapoints:
            calpha = d.binder_calpha()
            x.append(calpha[0])
            y.append(calpha[1])
            z.append(calpha[2])
            orient = d.binder_orientation()
            u.append(orient[0])
            v.append(orient[1])
            w.append(orient[2])
            if draw_secor:
                secor = d.binder_secondary_orientation()
                u2.append(secor[0])
                v2.append(secor[1])
                w2.append(secor[2])
        # draw the C_alpha -> C_beta vectors of the residues
        ax.quiver(x, y, z, u, v, w, color=col, arrow_length_ratio=0.5, linewidth=1.0)
        if draw_secor:
            # draw the C_alpha -> C_O vectors of the residues
            ax.quiver(x, y, z, u2, v2, w2, color=col, arrow_length_ratio=0.0, linewidth=1.0)
    plt.show()


def visualize_acomplex(acomplex: LigandBinderComplex, dimensions: float = 10.0, draw_secor: bool = True,
                       filter_score: bool = True, **score_args) -> None:
    """Shows a 3d quiver plot that contains the designed ligand (including backbone), the designed binder residues (in
    black) and the atlas datapoints associated with the corresponding ligand residue types.
    :param acomplex: the complex to visualize
    :param dimensions: sets the display limit of the plot. This spacing will be applied to every dimension of the
            outermost residue.
    :param draw_secor: if True, secondary orientation vectors are drawn (without arrowheads)
    :param filter_score: if True, only those aligned residues whose score is below 0.0 are displayed
    :param score_args: additional keyword arguments that will be passed on to prediction_matrix.proximity_score
    """
    # plot the data as matplotlib 3d quiver
    fig = plt.figure()
    plt.suptitle(acomplex.ligand.get_title())
    ax: Axes3D = fig.gca(projection='3d')
    min_dim = acomplex.ligand.get_min_dim()
    max_dim = acomplex.ligand.get_max_dim()
    ax.set_xlim(min_dim[0] - dimensions, max_dim[0] + dimensions)
    ax.set_ylim(min_dim[1] - dimensions, max_dim[1] + dimensions)
    ax.set_zlim(min_dim[2] - dimensions, max_dim[2] + dimensions)
    bbx: List[float] = []
    bby: List[float] = []
    bbz: List[float] = []
    bbu: List[float] = []
    bbv: List[float] = []
    bbw: List[float] = []
    bbprev: DesignedLigandResidue = None
    # process ligand's backbone-residue structure
    for ligres in acomplex.ligand.residues:
        # create an internal coordinate system for the considered residue
        licoor = get_icoor_from_aresidue(ligres)
        # show C_alpha->C_beta vector of each ligand residue
        lca = licoor.internal_to_external(Vector(0.0, 0.0, 0.0), False)
        lor = licoor.internal_to_external(Vector(4.5, 0.0, 0.0), True)
        # draw ligand residue
        ax.quiver([lca[0]], [lca[1]], [lca[2]], [lor[0]], [lor[1]], [lor[2]],
                  color=get_color(ligres.restype), arrow_length_ratio=0.4, linewidth=9.0)
        # fill backbone info
        if bbprev is not None:
            bbx.append(ligres.calpha()[0])
            bby.append(ligres.calpha()[1])
            bbz.append(ligres.calpha()[2])
            bbu.append(bbprev.calpha()[0] - ligres.calpha()[0])
            bbv.append(bbprev.calpha()[1] - ligres.calpha()[1])
            bbw.append(bbprev.calpha()[2] - ligres.calpha()[2])
        bbprev = ligres
    # draw backbone
    ax.quiver(bbx, bby, bbz, bbu, bbv, bbw, color="#000000", arrow_length_ratio=0.0, linewidth=9.0)
    # process binder's designed residue positions
    for binres in acomplex.binder.residues:
        # show C_alpha->C_beta vector of each designed binder residue
        bca = binres.calpha()
        bor = binres.orientation() ** 3.0
        # draw binder residue
        ax.quiver([bca[0]], [bca[1]], [bca[2]], [bor[0]], [bor[1]], [bor[2]],
                  color="#606060", arrow_length_ratio=0.7, linewidth=9.0)
        if draw_secor:
            bso = binres.secondary_orientation() ** 3.0
            ax.quiver([bca[0]], [bca[1]], [bca[2]], [bso[0]], [bso[1]], [bso[2]],
                      color="#606060", arrow_length_ratio=0.0, linewidth=9.0)
    # process the residues of the atlas
    for binrestype in color_map:
        # for each potential interacting binder residue, paint the quivers in
        # individual colors.
        relevant_datapoints = [d for d in acomplex.aligned_residues if d.binder_restype == binrestype]
        col = color_map[binrestype]
        x: List[float] = []
        y: List[float] = []
        z: List[float] = []
        u: List[float] = []
        v: List[float] = []
        w: List[float] = []
        u2: List[float] = []
        v2: List[float] = []
        w2: List[float] = []
        for d in relevant_datapoints:
            for binder_res in acomplex.binder.residues:
                score = proximity_score(binder_res, d, **score_args)
                if not filter_score or score > 0.0:
                    orient = d.orientation()
                    secor = d.secondary_orientation()
                    x.append(d.calpha()[0])
                    y.append(d.calpha()[1])
                    z.append(d.calpha()[2])
                    u.append(orient[0])
                    v.append(orient[1])
                    w.append(orient[2])
                    if draw_secor:
                        u2.append(secor[0])
                        v2.append(secor[1])
                        w2.append(secor[2])
        # draw the C_alpha -> C_beta vectors of the residues
        ax.quiver(x, y, z, u, v, w, color=col, arrow_length_ratio=0.5, linewidth=1.0)
        if draw_secor:
            # draw the C_alpha -> C_O vectors of the residues
            ax.quiver(x, y, z, u2, v2, w2, color=col, arrow_length_ratio=0.0, linewidth=1.0)
    plt.show()


def visualize_pmatrix(pmatrix: BinderPredictionMatrix) -> None:
    """Visualizes a given binder predictrion matrix as a stacked bar plot. Each stacked bar represents a residue of the
    designed binder. The sub-bars represent potential target residues each, where the type is coded by the bar color,
    the height encodes the score, and the width encodes the certainty.
    :param pmatrix: the prediction matrix to visualize
    """
    maxwidth = max(1, max([max([col.certainty for col in pline.columns], default=1) for pline in pmatrix.lines]))
    for ind, pline in enumerate(pmatrix.lines):
        bot = 0.0
        for pcolumn in pline.columns:
            col = get_color(pcolumn.residue_type)
            # height of the bar reflects score, width n_origins.
            plt.bar([ind], [pcolumn.score*100.0], pcolumn.certainty*(0.9/float(maxwidth)), color=col, bottom=[bot])
            bot += pcolumn.score*100.0
        # plot bars for reference amino acids
        reference_col = get_color(pline.original_restype)
        plt.bar([ind], 10.0, 0.9, color=reference_col, bottom=[-20])
    plt.xticks(range(0, len(pmatrix.lines)),
               [pline.original_restype + str(pline.residue_id) for pline in pmatrix.lines], rotation=60)
    plt.yticks([-15] + list(range(0, 100, 10)), ["Reference"] + list([f"{i}%" for i in range(0, 100, 10)]))
    plt.show()


def visualize_atlas_stats(atlas: Atlas, method: str = "plotly", **kwargs) -> None:
    """Visualizes the datapoints of an atlas as a stacked bar plot. Every stack represents a ligand residue type, and
    the individual bars represent the interacting binder residue types.

    :param atlas: the atlas to visualize
    :param method: Either "plotly" or "matplotlib". If not given plotly matplotlib is used.
    """
    if method == "plotly":
        return visualize_atlas_stats_plotly(atlas=atlas, **kwargs)
    else:
        return visualize_atlas_stats_matplotlib(atlas=atlas, **kwargs)


def visualize_atlas_stats_plotly(atlas: Atlas, ligand_per_binder: bool = False, color_mode: str = "0") -> None:
    """Visualizes the datapoints of an atlas as a stacked bar plot. Every stack represents a ligand residue type, and
    the individual bars represent the interacting binder residue types.
    param atlas: The atlas to take the data from
    param invert: if True, the plot maps ligand (x axis) to binder (y axis), else vice versa.
    """
    # TODO alternative amino acids
    stats = atlas.get_stats(ligand_per_binder=ligand_per_binder)
    stats = dict(sorted(Atlas.fill_stats(stats).items()))

    df = pd.DataFrame.from_dict(stats)
    df = df.transpose()
    cmap = get_color_map(color_mode)
    fig = px.bar(df,
                 x=df.index.name,
                 y=canonical_amino_acids,
                 color_discrete_map=cmap,
                 title="Wide-Form Input")

    fig.update_layout(
        title="Atlas statistics",
        xaxis_title="ligand amino acid" if not ligand_per_binder else "binder amino acid",
        yaxis_title="ligand amino acid" if ligand_per_binder else "binder amino acid")
    fig.show()


def visualize_atlas_stats_matplotlib(atlas: Atlas, ligand_per_binder: bool = False, color_mode: str = "0") -> None:
    """Visualizes the datapoints of an atlas as a stacked bar plot. Every stack represents a ligand residue type, and
    the individual bars represent the interacting binder residue types.
    param atlas: The atlas to take the data from
    param ligand_per_binder: if True, inverts the plot: ligand residues (y axis) per binder residues (x axis),
     else vice versa.
    """
    # TODO viualize alternative aas too...
    stats = atlas.get_stats(ligand_per_binder)
    stats = dict(sorted(Atlas.fill_stats(stats).items()))
    max_height = max([sum([val for val in line.values()]) for line in stats.values()])
    for ind, ligres in enumerate(canonical_amino_acids):
        if ligres not in stats:
            continue
        bot = 0.0
        for binres in sorted(canonical_amino_acids, key=lambda aa: stats[ligres].get(aa, 0.0), reverse=True):
            if binres not in stats[ligres]:
                continue
            count = stats[ligres][binres]
            height = float(count) / float(max_height) * 100.0
            bincol = get_color(binres, color_mode)
            plt.bar([ind], [height], 0.8, color=bincol, bottom=[bot])
            bot += height
        ligand_col = get_color(ligres, color_mode)
        plt.bar([ind], 10.0, 0.9, color=ligand_col, bottom=[-20.0])
    plt.xticks(range(0, 20), canonical_amino_acids, rotation=60)
    if max_height >= 100:
        plt.yticks([-15] + list(range(0, 101, 10)),
                   ["Binder" if ligand_per_binder else "Ligand"] +
                   list("{:.0f}".format(i / 10) for i in range(0, max_height * 11, max_height)))
    elif max_height:
        if max_height >= 50:
            divisor = 10
        elif max_height >= 15:
            divisor = 5
        else:
            divisor = 1
        # the next multiple of divisor greater than max_heigth
        next_multiple = int(max_height / divisor) + 1
        plt.yticks([-15] + list(round(divisor * 100 / max_height * i) for i in range(0, next_multiple + 1)),
                   ["Binder" if ligand_per_binder else "Ligand"] +
                   list(str(i) for i in range(0, next_multiple * divisor + 1, divisor)))
    plt.show()


def visualize_pocket_atlas(pocket: Pocket, method: str = "plotly", **kwargs):
    """Shows the datapoints of a pocket atlas with a plotly plot (within web broswer) or matplotlib (separate window).
        For additional parameters such as 'ligand_restype' compare the subfunctions used here.

        :param pocket: the pocket to visualize
        :param method: Either "plotly" or "matplotlib". If not plotly matplotlib is used.
    """
    return visualize_atlas(atlas=pocket.to_atlas(), method=method, ligand_restype=pocket.restype, **kwargs)


class DatapointNotFound(Exception):
    """raised if datapoint is not found in atlas."""

    def __init__(self, message="The datapoint you entered cannot be found within this atlas!"):
        self.message = message
        super().__init__(self.message)


def get_best_connections_for_dp(atoms_: List[Tuple[str, Vector, str]], restype_: str, trace_name: str, color_: str) \
       -> Iterator[Dict]:
    """
    Returns an Iterator including a dict with a trace for plotting (plotly?) bonds between the given atoms.
    Only works for canonical amino acids! # TODO non-canonical typical ones ?
    The backbone atoms will be grey, the sidechain atoms will be the assigned color.

    :param atoms_: The atoms that should be connected
    :param restype_: The residue type the atoms belong to
    :param trace_name: The name of the returned trace
    :param color_: The color of sidechain atoms
    :return: Iterator of Dicts including information about the trace to plot
    """

    def get_lines(canonical: bool):
        if not canonical:
            for bond in get_connected_atom_pairs(atoms_ + [("CA", Vector(0, 0, 0))]):
                yield {'type': "scatter3d", 'mode': 'lines',
                       'name': trace_name,
                       'x': [bond[0][0], bond[1][0]], 'y': [bond[0][1], bond[1][1]],
                       'z': [bond[0][2], bond[1][2]],
                       'line': {'width': 20, 'color': color_}}
        else:
            # For every connected line safed in best_connection define a bond of vectors
            for connected_atom_names in best_connection:
                bond_vectors = []
                if not len(connected_atom_names):
                    continue

                # Look for right atom names to match the predefined atoms in a line
                for connected_atom in connected_atom_names:
                    for (atom_name, atom_vector, origin) in atoms_:
                        # If name is found extend the connected line with vector
                        if atom_name == connected_atom:
                            bond_vectors.append(atom_vector)
                            break

                # TODO make warning? Assertion fails it atoms are missing in the input structure
                #  (e.g. Ser w/o OG - 3jb9 i/S9)
                # assert len(connected_atom_names) == len(bond_vectors)

                current_color = "#BBBBBB" if connected_atom_names[0] == "N" else color_

                yield {'type': "scatter3d", 'mode': 'lines', 'hoverinfo': 'none', 'name': trace_name,
                       'x': [b[0] for b in bond_vectors],
                       'y': [b[1] for b in bond_vectors], 'z': [b[2] for b in bond_vectors],
                       'line': {'width': 20, 'color': current_color}}

    # Set residue types and get atoms of residue
    try:
        # For canonical amino acids, the lines are predefined to use less traces (faster)
        best_connection = [("N", "CA", "C", "O"), connections[restype_]]
    except KeyError:
        return get_lines(canonical=False)

    return get_lines(canonical=True)


def get_connected_atom_pairs(all_atoms: Union[List[Tuple[str, Vector, str]], List[Tuple[str, Vector]]]):
    for a1, a2 in combinations(all_atoms, 2):
        if 0.5 < (a1[1] - a2[1]).norm() < 2:
            yield a1[1], a2[1]


def visualize_single_pocket(pocket: Pocket, datapoint: AtlasDatapoint, color_by_element: bool = False):

    # Filter pocket atlas for those who share the ligand origin with the datapoint parameter
    filtered_atlas = Atlas([dp for dp in pocket.to_atlas().datapoints if dp.ligand_origin == datapoint.ligand_origin])
    if not len(filtered_atlas.datapoints):
        raise DatapointNotFound
    dps = filtered_atlas.datapoints

    # define special (non-C) heavy atoms and their color
    hatom_names = ("N", "S", "P", "O")
    hatom_atoms = dict(zip(hatom_names, (defaultdict(list) for _ in hatom_names)))
    hatoms_colours = dict(zip(hatom_names, ("#0000DD", "#DDDD00", "#AB0000", "#DD0000")))

    # collect all coordinates, residue types and the size of the marker datapoints in lists.
    x = []
    y = []
    z = []
    restypes = []
    size = []
    bonds_list = []
    text = []

    # Set ligand atoms
    # Atom bubbles
    ligand_atoms: List[Tuple[str, Vector, str]] = [(name, atom, dps[0].ligand_origin)
                                                   for name, atom in dps[0].ligand_atoms.items()]
    ligand_restype = dps[0].ligand_restype
    for atom_name, atom_vector, origin in ligand_atoms:
        if not (color_by_element and atom_name[0] in hatom_names):
            x.append(atom_vector[0])
            y.append(atom_vector[1])
            z.append(atom_vector[2])
            restypes.append(ligand_restype)
            size.append(10)
            text.append(origin + "/" + atom_name)
        else:
            hatom_atoms[atom_name[0]]['x'].append(atom_vector[0])
            hatom_atoms[atom_name[0]]['y'].append(atom_vector[1])
            hatom_atoms[atom_name[0]]['z'].append(atom_vector[2])
            hatom_atoms[atom_name[0]]['text'].append(origin + "/" + atom_name)

    for ligand_line in get_best_connections_for_dp(ligand_atoms, restype_=ligand_restype,
                                                   trace_name=ligand_restype + " ligand bond",
                                                   color_=get_color(ligand_restype)):
        bonds_list.append(ligand_line)

    # For every datapoint save coords, restype and size (Ca and Cb)
    for dp in dps:
        bin_atoms = []

        # Define atom bubbles
        for atom_name, atom_vector in dp.binder_atoms.items():
            bin_atoms.append((atom_name, atom_vector, dp.binder_origin))
            # if color_by_element, save heavy atoms differently to plot different atom colours.
            if not (color_by_element and atom_name[0] in hatom_names):
                x.append(atom_vector[0])
                y.append(atom_vector[1])
                z.append(atom_vector[2])
                restypes.append(dp.binder_restype)
                size.append(8)
                text.append(dp.binder_origin + "/" + atom_name)
            else:
                hatom_atoms[atom_name[0]]['x'].append(atom_vector[0])
                hatom_atoms[atom_name[0]]['y'].append(atom_vector[1])
                hatom_atoms[atom_name[0]]['z'].append(atom_vector[2])
                hatom_atoms[atom_name[0]]['text'].append(dp.binder_origin + "/" + atom_name)

        # define connections between residue atoms for plotting of covalent bonds.
        for binder_line in get_best_connections_for_dp(bin_atoms, restype_=dp.binder_restype,
                                                       trace_name=dp.binder_restype + " bond",
                                                       color_=get_color(dp.binder_restype)):
            bonds_list.append(binder_line)

    # Define dataframe for markers of binder atoms
    df = pd.DataFrame.from_dict({
        "x": x,
        "y": y,
        "z": z,
        "restype": restypes,
        "size": size,
        "text": text,
    })

    # concatenate color dictionaries to address Ca (full color) and Cb (light color)
    color_map_light_cb = {k + " cb": v for k, v in color_map_light.items()}
    color_map_full = {**color_map, **color_map_light_cb}

    # Define figure and include all atoms (if color_by_element: only carbon atoms) as markers
    fig = px.scatter_3d(df, x='x', y='y', z='z', hover_name='text',
                        color='restype', color_discrete_map=color_map_full)

    # Special heavy atom coloring: O, N, S, P as markers
    if color_by_element:
        for atoms_name, atoms in hatom_atoms.items():
            fig.add_trace(go.Scatter3d(x=atoms['x'], y=atoms['y'], z=atoms['z'], name=atoms_name,
                                       hovertemplate="<b>%{hovertext}</b><br><br>x=%{x}<br>y=%{y}<br>z=%{z}",
                                       mode='markers', showlegend=False, hovertext=atoms['text'],
                                       marker={'line': {'width': 20, 'color': hatoms_colours[atoms_name]}}))

    # Define bond lines
    for bond in bonds_list:
        fig.add_trace(go.Scatter3d(x=bond['x'], y=bond['y'], z=bond['z'], name=bond['name'], mode='lines',
                                   showlegend=False, line=bond['line'], hoverinfo='skip'))
    fig.update_layout(
        scene=dict(
            xaxis=dict(range=[-15, 15], ),
            yaxis=dict(range=[-15, 15], ),
            zaxis=dict(range=[-15, 15], ),
            aspectmode="cube", ))
    fig.show()
